#!/usr/bin/env python
# Copyright 2014-2021 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

import sys
from functools import reduce
import numpy, scipy
from pyscf import lib
from pyscf import gto
from pyscf.lib import logger
from pyscf import fci
from pyscf import scf
from pyscf import symm
from pyscf import __config__

BASE = getattr(__config__, 'mcscf_addons_sort_mo_base', 1)
MAP2HF_TOL = getattr(__config__, 'mcscf_addons_map2hf_tol', 0.4)

if sys.version_info < (3,):
    RANGE_TYPE = list
else:
    RANGE_TYPE = range


def sort_mo(casscf, mo_coeff, caslst, base=BASE):
    '''Pick orbitals for CAS space

    Args:
        casscf : an :class:`CASSCF` or :class:`CASCI` object

        mo_coeff : ndarray or a list of ndarray
            Orbitals for CASSCF initial guess.  In the UHF-CASSCF, it's a list
            of two orbitals, for alpha and beta spin.
        caslst : list of int or nested list of int
            A list of orbital indices to represent the CAS space.  In the UHF-CASSCF,
            it's consist of two lists, for alpha and beta spin.

    Kwargs:
        base : int
            0-based (C-style) or 1-based (Fortran-style) caslst

    Returns:
        An reoreded mo_coeff, which put the orbitals given by caslst in the CAS space

    Examples:

    >>> from pyscf import gto, scf, mcscf
    >>> mol = gto.M(atom='N 0 0 0; N 0 0 1', basis='ccpvdz', verbose=0)
    >>> mf = scf.RHF(mol)
    >>> mf.scf()
    >>> mc = mcscf.CASSCF(mf, 4, 4)
    >>> cas_list = [5,6,8,9] # pi orbitals
    >>> mo = mc.sort_mo(cas_list)
    >>> mc.kernel(mo)[0]
    -109.007378939813691
    '''
    ncore = casscf.ncore

    def ext_list(nmo, caslst):
        mask = numpy.ones(nmo, dtype=bool)
        mask[caslst] = False
        idx = numpy.where(mask)[0]
        if len(idx) + casscf.ncas != nmo:
            raise ValueError('Active space size is incompatible with caslist. '
                             'ncas = %d.  caslist %s' % (casscf.ncas, caslst))
        return idx

    if isinstance(ncore, (int, numpy.integer)):
        nmo = mo_coeff.shape[1]
        if base != 0:
            caslst = [i-base for i in caslst]
        idx = ext_list(nmo, caslst)
        mo = numpy.hstack((mo_coeff[:,idx[:ncore]],
                           mo_coeff[:,caslst],
                           mo_coeff[:,idx[ncore:]]))

        if getattr(mo_coeff, 'orbsym', None) is not None:
            orbsym = mo_coeff.orbsym
            orbsym = numpy.hstack((orbsym[idx[:ncore]], orbsym[caslst],
                                   orbsym[idx[ncore:]]))
            mo = lib.tag_array(mo, orbsym=orbsym)
        return mo

    else: # UHF-based CASSCF
        if isinstance(caslst[0], (int, numpy.integer)):
            if base != 0:
                caslsta = [i-1 for i in caslst]
                caslst = (caslsta, caslsta)
        else: # two casspace lists, for alpha and beta
            if base != 0:
                caslst = ([i-base for i in caslst[0]],
                          [i-base for i in caslst[1]])
        nmo = mo_coeff[0].shape[1]
        idxa = ext_list(nmo, caslst[0])
        mo_a = numpy.hstack((mo_coeff[0][:,idxa[:ncore[0]]],
                             mo_coeff[0][:,caslst[0]],
                             mo_coeff[0][:,idxa[ncore[0]:]]))
        idxb = ext_list(nmo, caslst[1])
        mo_b = numpy.hstack((mo_coeff[1][:,idxb[:ncore[1]]],
                             mo_coeff[1][:,caslst[1]],
                             mo_coeff[1][:,idxb[ncore[1]:]]))

        if getattr(mo_coeff[0], 'orbsym', None) is not None:
            orbsyma, orbsymb = mo_coeff[0].orbsym, mo_coeff[1].orbsym
            orbsyma = numpy.hstack((orbsyma[idxa[:ncore[0]]], orbsyma[caslst[0]],
                                    orbsyma[idxa[ncore[0]:]]))
            orbsymb = numpy.hstack((orbsymb[idxb[:ncore[1]]], orbsymb[caslst[1]],
                                    orbsymb[idxb[ncore[1]:]]))
            mo_a = lib.tag_array(mo_a, orbsym=orbsyma)
            mo_b = lib.tag_array(mo_b, orbsym=orbsymb)
        return (mo_a, mo_b)

def select_mo_by_irrep(casscf,  cas_occ_num, mo=None, base=BASE):
    raise RuntimeError('This function has been replaced by function caslst_by_irrep')

def caslst_by_irrep(casscf, mo_coeff, cas_irrep_nocc,
                    cas_irrep_ncore=None, s=None, base=BASE):
    '''Given number of active orbitals for each irrep, return the orbital
    indices of active space

    Args:
        casscf : an :class:`CASSCF` or :class:`CASCI` object

        cas_irrep_nocc : list or dict
            Number of active orbitals for each irrep.  It can be a dict, eg
            {'A1': 2, 'B2': 4} to indicate the active space size based on
            irrep names, or {0: 2, 3: 4} for irrep Id,  or a list [2, 0, 0, 4]
            (identical to {0: 2, 3: 4}) in which the list index is served as
            the irrep Id.

    Kwargs:
        cas_irrep_ncore : list or dict
            Number of closed shells for each irrep.  It can be a dict, eg
            {'A1': 6, 'B2': 4} to indicate the closed shells based on
            irrep names, or {0: 6, 3: 4} for irrep Id,  or a list [6, 0, 0, 4]
            (identical to {0: 6, 3: 4}) in which the list index is served as
            the irrep Id.  If cas_irrep_ncore is not given, the program
            will generate a guess based on the lowest :attr:`CASCI.ncore`
            orbitals.
        s : ndarray
            overlap matrix
        base : int
            0-based (C-like) or 1-based (Fortran-like) caslst

    Returns:
        A list of orbital indices

    Examples:

    >>> from pyscf import gto, scf, mcscf
    >>> mol = gto.M(atom='N 0 0 0; N 0 0 1', basis='ccpvtz', symmetry=True, verbose=0)
    >>> mf = scf.RHF(mol)
    >>> mf.kernel()
    >>> mc = mcscf.CASSCF(mf, 12, 4)
    >>> mcscf.caslst_by_irrep(mc, mf.mo_coeff, {'E1gx':4, 'E1gy':4, 'E1ux':2, 'E1uy':2})
    [5, 7, 8, 10, 11, 14, 15, 20, 25, 26, 31, 32]
    '''
    mol = casscf.mol
    log = logger.Logger(casscf.stdout, casscf.verbose)
    orbsym = numpy.asarray(scf.hf_symm.get_orbsym(mol, mo_coeff))
    ncore = casscf.ncore

    irreps = set(orbsym)

    if cas_irrep_ncore is not None:
        irrep_ncore = {}
        for k, v in cas_irrep_ncore.items():
            if isinstance(k, str):
                irrep_ncore[symm.irrep_name2id(mol.groupname, k)] = v
            else:
                irrep_ncore[k] = v

        ncore_rest = ncore - sum(irrep_ncore.values())
        if ncore_rest > 0:  # guess core configuration
            mask = numpy.ones(len(orbsym), dtype=bool)
            for ir in irrep_ncore:
                mask[orbsym == ir] = False
            core_rest = orbsym[mask][:ncore_rest]
            core_rest = dict([(ir, numpy.count_nonzero(core_rest==ir))
                              for ir in set(core_rest)])
            log.info('Given core space %s < casscf core size %d',
                     cas_irrep_ncore, ncore)
            log.info('Add %s to core configuration', core_rest)
            irrep_ncore.update(core_rest)
        elif ncore_rest < 0:
            raise ValueError('Given core space %s > casscf core size %d'
                             % (cas_irrep_ncore, ncore))
    else:
        irrep_ncore = dict([(ir, sum(orbsym[:ncore]==ir)) for ir in irreps])

    if not isinstance(cas_irrep_nocc, dict):
        # list => dict
        cas_irrep_nocc = dict([(ir, n) for ir,n in enumerate(cas_irrep_nocc)
                               if n > 0])

    irrep_ncas = {}
    for k, v in cas_irrep_nocc.items():
        if isinstance(k, str):
            irrep_ncas[symm.irrep_name2id(mol.groupname, k)] = v
        else:
            irrep_ncas[k] = v

    ncas_rest = casscf.ncas - sum(irrep_ncas.values())
    if ncas_rest > 0:
        mask = numpy.ones(len(orbsym), dtype=bool)
# remove core and specified active space
        for ir in irrep_ncas:
            mask[orbsym == ir] = False
        for ir, ncore in irrep_ncore.items():
            idx = numpy.where(orbsym == ir)[0]
            mask[idx[:ncore]] = False

        cas_rest = orbsym[mask][:ncas_rest]
        cas_rest = dict([(ir, numpy.count_nonzero(cas_rest==ir))
                         for ir in set(cas_rest)])
        log.info('Given active space %s < casscf active space size %d',
                 cas_irrep_nocc, casscf.ncas)
        log.info('Add %s to active space', cas_rest)
        irrep_ncas.update(cas_rest)
    elif ncas_rest < 0:
        raise ValueError('Given active space %s > casscf active space size %d'
                         % (cas_irrep_nocc, casscf.ncas))

    caslst = []
    for ir, ncas in irrep_ncas.items():
        if ncas > 0:
            if ir in irrep_ncore:
                nc = irrep_ncore[ir]
            else:
                nc = 0
            no = nc + ncas
            idx = numpy.where(orbsym == ir)[0]
            caslst.extend(idx[nc:no])
    caslst = numpy.sort(numpy.asarray(caslst)) + base
    if len(caslst) < casscf.ncas:
        raise ValueError('Not enough orbitals found for core %s, cas %s' %
                         (cas_irrep_ncore, cas_irrep_nocc))

    if log.verbose >= logger.INFO:
        log.info('ncore for each irreps %s',
                 dict([(symm.irrep_id2name(mol.groupname, k), v)
                       for k,v in irrep_ncore.items()]))
        log.info('ncas for each irreps %s',
                 dict([(symm.irrep_id2name(mol.groupname, k), v)
                       for k,v in irrep_ncas.items()]))
        log.info('(%d-based) caslst = %s', base, caslst)
    return caslst

def sort_mo_by_irrep(casscf, mo_coeff, cas_irrep_nocc,
                     cas_irrep_ncore=None, s=None):
    '''Given number of active orbitals for each irrep, construct the mo initial
    guess for CASSCF

    Args:
        casscf : an :class:`CASSCF` or :class:`CASCI` object

        cas_irrep_nocc : list or dict
            Number of active orbitals for each irrep.  It can be a dict, eg
            {'A1': 2, 'B2': 4} to indicate the active space size based on
            irrep names, or {0: 2, 3: 4} for irrep Id,  or a list [2, 0, 0, 4]
            (identical to {0: 2, 3: 4}) in which the list index is served as
            the irrep Id.

    Kwargs:
        cas_irrep_ncore : list or dict
            Number of closed shells for each irrep.  It can be a dict, eg
            {'A1': 6, 'B2': 4} to indicate the closed shells based on
            irrep names, or {0: 6, 3: 4} for irrep Id,  or a list [6, 0, 0, 4]
            (identical to {0: 6, 3: 4}) in which the list index is served as
            the irrep Id.  If cas_irrep_ncore is not given, the program
            will generate a guess based on the lowest :attr:`CASCI.ncore`
            orbitals.
        s : ndarray
            overlap matrix

    Returns:
        sorted orbitals, ordered as [c,..,c,a,..,a,v,..,v]

    Examples:

    >>> from pyscf import gto, scf, mcscf
    >>> mol = gto.M(atom='N 0 0 0; N 0 0 1', basis='ccpvtz', symmetry=True, verbose=0)
    >>> mf = scf.RHF(mol)
    >>> mf.kernel()
    >>> mc = mcscf.CASSCF(mf, 12, 4)
    >>> mo = mc.sort_mo_by_irrep({'E1gx':4, 'E1gy':4, 'E1ux':2, 'E1uy':2})
    >>> # Same to mo = sort_mo_by_irrep(mc, mf.mo_coeff, {2: 4, 3: 4, 6: 2, 7: 2})
    >>> # Same to mo = sort_mo_by_irrep(mc, mf.mo_coeff, [0, 0, 4, 4, 0, 0, 2, 2])
    >>> mc.kernel(mo)[0]
    -108.162863845084
    '''
    caslst = caslst_by_irrep(casscf, mo_coeff, cas_irrep_nocc,
                             cas_irrep_ncore, s, base=0)
    return sort_mo(casscf, mo_coeff, caslst, base=0)

def make_natural_orbitals (method_obj):
    """Make natural orbitals from a general PySCF method object.
    See Eqn. (1) in Keller et al. [DOI:10.1063/1.4922352] for details.


    Args:
        method_obj : Any PySCF method that has the function `make_rdm1` with kwarg
            `ao_repr`. This object can be a restricted OR unrestricted method.

    Returns:
        noons : A 1-D array of the natural orbital occupations (NOONs).

        natorbs : A set of natural orbitals made from method_obj.
    """
    mf = method_obj
    if hasattr(method_obj, "_scf"):
        mf = method_obj._scf
    rdm1 = method_obj.make_rdm1(ao_repr=True)
    S = mf.get_ovlp()

    # Slight difference for restricted vs. unrestriced case
    if isinstance(rdm1, tuple):
        Dm = rdm1[0]+rdm1[1]
    elif isinstance(rdm1, numpy.ndarray):
        if numpy.ndim(rdm1) == 3:
            Dm = rdm1[0]+rdm1[1]
        elif numpy.ndim(rdm1) == 2:
            Dm = rdm1
        else:
            raise ValueError(
                "rdm1 passed to is a numpy array," +
                "but it has the wrong number of dimensions: {}".format(numpy.ndim(rdm1)))
    else:
        raise ValueError(
            "\n\tThe rdm1 generated by method_obj.make_rdm1() was a {}."
            "\n\tThis type is not supported, please select a different method and/or "
            "open an issue at https://github.com/pyscf/pyscf/issues".format(type(rdm1))
        )

    # Diagonalize the DM in AO (using Eqn. (1) referenced above)
    A = reduce(numpy.dot, (S, Dm, S))
    w, v = scipy.linalg.eigh(A, b=S)

    # Flip NOONs (and NOs) since they're in increasing order
    noons = numpy.flip(w)
    natorbs = numpy.flip(v, axis=1)

    return noons, natorbs


def project_init_guess (casscf, mo_init, prev_mol=None, priority=None, use_hf_core=None):
    '''Project the given initial guess to the current CASSCF problem
    giving using a sequence of SVDs on orthogonal orbital subspaces.

    Args:
        casscf : an :class:`CASSCF` or :class:`CASCI` object

        mo_init : ndarray or list of ndarray
            Initial guess orbitals which are not orthonormal for the
            current molecule.  When the casscf is UHF-CASSCF, mo_init
            needs to be a list of two ndarrays, for alpha and beta
            orbitals. Cannot have linear dependencies (i.e., you cannot
            give more orbitals than the basis of casscf.mol has). Must
            have at least ncore+ncas columns with active orbitals last,
            even if use_hf_core=True. If incomplete, additional virtual
            orbitals will be constructed and appended automatically.

    Kwargs:
        prev_mol : an instance of :class:`Mole`
            If given, the inital guess orbitals are associated to the
            basis of prev_mol. Otherwise, the orbitals are presumed to
            be in the basis of casscf.mol. Beware linear dependencies if
            you are projecting from a LARGER basis to a SMALLER one.

        priority : 'active', 'core', nested idx arrays, or mask array
            If arrays are 3d, UHF-CASSCF must be used; arrays can always
            be 2d. Specifies the order in which groups of orbitals are
            projected. Orbitals orthogonalized earlier are deformed less
            than those orthogonalized later. 'core' means core, then
            active, then virtual; 'active' means active, then core, then
            virtual, and the Gram-Schmidt process is generated by
            [[0],[1],[2],...] or numpy.eye (nmo). Missing orbitals are
            presumed virtual. Defaults to 'active' if you are projecting
            from the same basis set (prev_mol is None or has the same
            basis functions) and 'core' otherwise.

        use_hf_core : logical
            If True, the core orbitals of mo_init are swapped out with
            HF orbitals chosen by maximum overlap. Defaults to True
            if you are projecting from a different basis and False if
            you are projecting from a different geometry.

    Returns:
        New orthonormal initial guess orbitals'''

    from pyscf import lo
    ncore, ncas = casscf.ncore, casscf.ncas
    s0 = casscf._scf.get_ovlp ()
    mf_mo = casscf._scf.mo_coeff
    nmo = numpy.asarray (mf_mo).shape[-1]
    nmo_init = numpy.asarray (mo_init).shape[-1]
    if nmo_init > nmo:
        raise RuntimeError ("Too many orbitals in mo_init (try passing only the occupied orbitals)")

    # Project orbitals from a different basis
    if prev_mol is not None:
        if gto.same_mol(prev_mol, casscf.mol, cmp_basis=False):
            if isinstance(ncore, (int, numpy.integer)):  # RHF
                mo_init = scf.addons.project_mo_nr2nr(prev_mol, mo_init, casscf.mol)
            else:
                mo_init = (scf.addons.project_mo_nr2nr(prev_mol, mo_init[0], casscf.mol),
                           scf.addons.project_mo_nr2nr(prev_mol, mo_init[1], casscf.mol))
        elif gto.same_basis_set(prev_mol, casscf.mol):
            prev_mol = None # Ignore! Serves no purpose unless it's a different basis set
        else:
            raise NotImplementedError('Project initial guess from different system.')
    if priority is None: priority = ('core','active')[prev_mol is None]
    if use_hf_core is None: use_hf_core = (prev_mol is not None)

    # Do the projection
    def _symmcase (mo_basis, mo_target):
        ovlp = reduce (numpy.dot, (mo_basis.conj ().T, s0, mo_target))
        u, s, vh = scipy.linalg.svd (ovlp, full_matrices=True)
        mo_proj = reduce (numpy.dot, (mo_basis, u[:,:len(s)], vh))
        if u.shape[1] > len(s):
            mo_null = numpy.dot (mo_basis, u[:,len(s):])
        else:
            mo_null = numpy.zeros ((mo_basis.shape[0], 0))
        return mo_proj, mo_null

    # Iterate over irreps
    def _rangecase (mo_basis, mo_target):
        if not casscf.mol.symmetry:
            return _symmcase (mo_basis, mo_target)
        orbsym_target = scf.hf_symm.get_orbsym(casscf.mol, mo_target, s0)
        orbsym_basis = scf.hf_symm.get_orbsym(casscf.mol, mo_basis, s0)
        for ir in set (orbsym_target):
            errstr = 'inadequate basis for symmetry {}'.format (ir)
            assert (numpy.count_nonzero (orbsym_basis==ir) >=
                    numpy.count_nonzero (orbsym_target==ir)), errstr
        # Avoid scrambling the order of the target orbitals
        mo_null = []
        mo_proj = numpy.zeros_like (mo_target)
        for ir in set (orbsym_basis):
            mo_basis_ir = mo_basis[:,orbsym_basis==ir]
            idx_ir = orbsym_target == ir
            if numpy.count_nonzero (idx_ir) == 0:
                mo_null.append (mo_basis_ir)
                continue
            mo_proj_ir, mo_null_ir = _symmcase (mo_basis_ir, mo_target[:,idx_ir])
            mo_null.append (mo_null_ir)
            mo_proj[:,idx_ir] = mo_proj_ir
        mo_null = numpy.hstack (mo_null)
        return mo_proj, mo_null

    # Iterate over orbital ranges
    def _spincase (mo_basis, mo_target, range_idx, sort_idx, ncore):
        # Swap out HF core orbitals (making a copy for safety)
        if use_hf_core:
            ovlp = numpy.dot (mo_target.conj ().T, s0.dot (mo_basis))
            ix = numpy.argmax (numpy.abs (ovlp[:ncore,:]), axis=1)
            mo_target = numpy.append (mo_basis[:,ix], mo_target[:,ncore:], axis=1)
        # Do the iteration
        mo = numpy.zeros_like (mo_target)
        for idx in range_idx:
            mo[:,idx], mo_basis = _rangecase (mo_basis, mo_target[:,idx])
        idx = numpy.any (range_idx, axis=0)
        mo = mo[:,idx]
        mo_target = mo_target[:,idx]
        # Fix sign
        sgn = numpy.einsum ('pi,pi->i', mo.conj (), s0.dot (mo_target))
        mo[:,sgn<0] *= -1
        # Append remaining virtual orbitals
        if mo_basis.shape[-1] > 0:
            mo = numpy.append (mo, mo_basis, axis=1)
        # Sort and debug print
        mo[:,:nmo_init] = mo[:,:nmo_init][:,sort_idx]
        if casscf.verbose >= logger.DEBUG:
            mocc = mo[:,:ncore+ncas]
            s1 = reduce(numpy.dot, (mocc.T, s0, mo_target))
            tnorm = numpy.einsum ('pi,pi->i', mo_target.conj (), s0.dot (mo_target))
            s1_norm = s1 / numpy.sqrt (tnorm) [None,:]
            idx = numpy.argmax (numpy.abs (s1), axis=1)
            for i, j in enumerate (idx):
                logger.debug(casscf, 'Init guess <mo-orth|mo-init>  %d  %d  %10.8f (%10.8f after norm)',
                             i+1, j+1, s1[i,j], s1_norm[i,j])
        return mo

    # Interpret "priority" keyword
    def _interpret (priority, ncore):
        # Interpret priority keyword
        nocc = ncore + ncas
        if isinstance (priority, str):
            ridx = numpy.zeros ((2, nmo_init), dtype=bool)
            ridx[0,:ncore] = ridx[1,ncore:nocc] = True
            if priority.lower () == 'active': ridx = ridx[::-1,:]
            elif not priority.lower () == 'core':
                raise RuntimeError ("Invalid priority keyword: string must be either 'active' or 'core'")
            # Edge case: ncore == 0 or ncas == 0 -> remove zero rows from ridx
            ridx = ridx[ridx.sum (1).astype (bool)]
        else:
            ridx = numpy.zeros ((len (priority), nmo), dtype=bool)
            for row, idx in zip (ridx, priority):
                try:
                    row[idx] = True
                except IndexError:
                    raise RuntimeError ("Invalid priority keyword: index array cannot address shape (*,nmo_init)")
            ridx_counts = ridx.astype (int).sum (0)
            if numpy.any (ridx_counts > 1):
                raise RuntimeError ("Invalid priority keyword: index array has repeated elements")
        incl = numpy.any (ridx, axis=0)
        sidx = numpy.append (numpy.where (incl)[0], numpy.where (~incl)[0])
        return ridx, numpy.argsort (sidx)

    # Iterate over spin cases
    if isinstance(ncore, (int, numpy.integer)):
        # errstr = 'Invalid priority keyword (3-dim is valid for UHF-CAS only)'
        range_idx, sort_idx = _interpret (priority, ncore)
        mo = _spincase (mf_mo, mo_init, range_idx, sort_idx, ncore)
    else: # UHF-based CASSCF
        if (isinstance (priority, str)  # single string
            or (isinstance (priority, numpy.ndarray) and priority.ndim == 2)  # single mask array
            or isinstance (priority[0][0], (int, numpy.integer))): # 2d nested list
            priority = [priority, priority]
        idx = [_interpret (p, n) for p, n in zip (priority, ncore)]
        mo = (_spincase (mf_mo[0], mo_init[0], idx[0][0], idx[0][1], ncore[0]),
              _spincase (mf_mo[1], mo_init[1], idx[1][0], idx[1][1], ncore[1]))

    return mo


def project_init_guess_old(casscf, init_mo, prev_mol=None):
    '''Project the given initial guess to the current CASSCF problem.  The
    projected initial guess has two parts.  The core orbitals are directly
    taken from the Hartree-Fock orbitals, and the active orbitals are
    projected from the given initial guess.

    Args:
        casscf : an :class:`CASSCF` or :class:`CASCI` object

        init_mo : ndarray or list of ndarray
            Initial guess orbitals which are not orth-normal for the current
            molecule.  When the casscf is UHF-CASSCF, the init_mo needs to be
            a list of two ndarrays, for alpha and beta orbitals

    Kwargs:
        prev_mol : an instance of :class:`Mole`
            If given, the inital guess orbitals are associated to the geometry
            and basis of prev_mol.  Otherwise, the orbitals are based of
            the geometry and basis of casscf.mol

    Returns:
        New orthogonal initial guess orbitals with the core taken from
        Hartree-Fock orbitals and projected active space from original initial
        guess orbitals

    Examples:

    .. code:: python

        import numpy
        from pyscf import gto, scf, mcscf
        mol = gto.Mole()
        mol.build(atom='H 0 0 0; F 0 0 0.8', basis='ccpvdz', verbose=0)
        mf = scf.RHF(mol)
        mf.scf()
        mc = mcscf.CASSCF(mf, 6, 6)
        mo = mcscf.sort_mo(mc, mf.mo_coeff, [3,4,5,6,8,9])
        print('E(0.8) = %.12f' % mc.kernel(mo)[0])
        init_mo = mc.mo_coeff
        for b in numpy.arange(1.0, 3., .2):
            mol.atom = [['H', (0, 0, 0)], ['F', (0, 0, b)]]
            mol.build(0, 0)
            mf = scf.RHF(mol)
            mf.scf()
            mc = mcscf.CASSCF(mf, 6, 6)
            mo = mcscf.project_init_guess(mc, init_mo)
            print('E(%2.1f) = %.12f' % (b, mc.kernel(mo)[0]))
            init_mo = mc.mo_coeff
    '''
    from pyscf import lo

    def project(mfmo, init_mo, ncore, s):
        s_init_mo = numpy.einsum('pi,pi->i', init_mo.conj(), s.dot(init_mo))
        if abs(s_init_mo - 1).max() < 1e-7 and mfmo.shape[1] == init_mo.shape[1]:
            # Initial guess orbitals are orthonormal
            return init_mo
# TODO: test whether the canonicalized orbitals are better than the projected orbitals
# Be careful that the ordering of the canonicalized orbitals may be very different
# to the CASSCF orbitals.
#        else:
#            fock = casscf.get_fock(mc, init_mo, casscf.ci)
#            return casscf._scf.eig(fock, s)[1]

        nocc = ncore + casscf.ncas
        if ncore > 0:
            mo0core = init_mo[:,:ncore]
            s1 = reduce(numpy.dot, (mfmo.T, s, mo0core))
            s1core = reduce(numpy.dot, (mo0core.T, s, mo0core))
            coreocc = numpy.einsum('ij,ji->i', s1, lib.cho_solve(s1core, s1.T))
            coreidx = numpy.sort(numpy.argsort(-coreocc)[:ncore])
            logger.debug(casscf, 'Core indices %s', coreidx)
            logger.debug(casscf, 'Core components %s', coreocc[coreidx])
            # take HF core
            mocore = mfmo[:,coreidx]

            # take projected CAS space
            mocas = init_mo[:,ncore:nocc] \
                  - reduce(numpy.dot, (mocore, mocore.T, s, init_mo[:,ncore:nocc]))
            mocc = lo.orth.vec_lowdin(numpy.hstack((mocore, mocas)), s)
        else:
            mocc = lo.orth.vec_lowdin(init_mo[:,:nocc], s)

        # remove core and active space from rest
        if mocc.shape[1] < mfmo.shape[1]:
            if casscf.mol.symmetry:
                restorb = []
                orbsym = scf.hf_symm.get_orbsym(casscf.mol, mfmo, s)
                for ir in set(orbsym):
                    mo_ir = mfmo[:,orbsym==ir]
                    rest = mo_ir - reduce(numpy.dot, (mocc, mocc.T, s, mo_ir))
                    e, u = numpy.linalg.eigh(reduce(numpy.dot, (rest.T, s, rest)))
                    restorb.append(numpy.dot(rest, u[:,e>1e-7]))
                restorb = numpy.hstack(restorb)
            else:
                rest = mfmo - reduce(numpy.dot, (mocc, mocc.T, s, mfmo))
                e, u = numpy.linalg.eigh(reduce(numpy.dot, (rest.T, s, rest)))
                restorb = numpy.dot(rest, u[:,e>1e-7])
            mo = numpy.hstack((mocc, restorb))
        else:
            mo = mocc

        if casscf.verbose >= logger.DEBUG:
            s1 = reduce(numpy.dot, (mo[:,ncore:nocc].T, s, mfmo))
            idx = numpy.argwhere(abs(s1) > 0.4)
            for i,j in idx:
                logger.debug(casscf, 'Init guess <mo-CAS|mo-hf>  %d  %d  %12.8f',
                             ncore+i+1, j+1, s1[i,j])
        return mo

    ncore = casscf.ncore
    mfmo = casscf._scf.mo_coeff
    s = casscf._scf.get_ovlp()
    if prev_mol is None:
        if init_mo.shape[0] != mfmo.shape[0]:
            raise RuntimeError('Initial guess orbitals has wrong dimension')
    elif gto.same_mol(prev_mol, casscf.mol, cmp_basis=False):
        if isinstance(ncore, (int, numpy.integer)):  # RHF
            init_mo = scf.addons.project_mo_nr2nr(prev_mol, init_mo, casscf.mol)
        else:
            init_mo = (scf.addons.project_mo_nr2nr(prev_mol, init_mo[0], casscf.mol),
                       scf.addons.project_mo_nr2nr(prev_mol, init_mo[1], casscf.mol))
    elif gto.same_basis_set(prev_mol, casscf.mol):
        if isinstance(ncore, (int, numpy.integer)):  # RHF
            fock = casscf.get_fock(init_mo, casscf.ci)
            return casscf._scf.eig(fock, s)[1]
        else:
            raise NotImplementedError('Project initial for UHF orbitals.')
    else:
        raise NotImplementedError('Project initial guess from different system.')

# Be careful with the orbital projection. The projection may lead to bad
# initial guess orbitals if the geometry is dramatically changed.
    if isinstance(ncore, (int, numpy.integer)):
        mo = project(mfmo, init_mo, ncore, s)
    else: # UHF-based CASSCF
        mo = (project(mfmo[0], init_mo[0], ncore[0], s),
              project(mfmo[1], init_mo[1], ncore[1], s))
    return mo

# on AO representation
def make_rdm1(casscf, mo_coeff=None, ci=None, **kwargs):
    '''One-particle densit matrix in AO representation

    Args:
        casscf : an :class:`CASSCF` or :class:`CASCI` object

    Kwargs:
        ci : ndarray
            CAS space FCI coefficients. If not given, take casscf.ci.
        mo_coeff : ndarray
            Orbital coefficients. If not given, take casscf.mo_coeff.

    Examples:

    >>> import scipy.linalg
    >>> from pyscf import gto, scf, mcscf
    >>> mol = gto.M(atom='N 0 0 0; N 0 0 1', basis='sto-3g', verbose=0)
    >>> mf = scf.RHF(mol)
    >>> res = mf.scf()
    >>> mc = mcscf.CASSCF(mf, 6, 6)
    >>> res = mc.kernel()
    >>> natocc = numpy.linalg.eigh(mcscf.make_rdm1(mc), mf.get_ovlp(), type=2)[0]
    >>> print(natocc)
    [ 0.0121563   0.0494735   0.0494735   1.95040395  1.95040395  1.98808879
      2.          2.          2.          2.        ]
    '''
    return casscf.make_rdm1(mo_coeff, ci, **kwargs)

# make both alpha and beta density matrices
def make_rdm1s(casscf, mo_coeff=None, ci=None, **kwargs):
    '''Alpha and beta one-particle densit matrices in AO representation
    '''
    return casscf.make_rdm1s(mo_coeff, ci, **kwargs)

def _is_uhf_mo(mo_coeff):
    return not (isinstance(mo_coeff, numpy.ndarray) and mo_coeff.ndim == 2)

def _make_rdm12_on_mo(casdm1, casdm2, ncore, ncas, nmo):
    nocc = ncas + ncore
    dm1 = numpy.zeros((nmo,nmo))
    idx = numpy.arange(ncore)
    dm1[idx,idx] = 2
    dm1[ncore:nocc,ncore:nocc] = casdm1

    dm2 = numpy.zeros((nmo,nmo,nmo,nmo))
    dm2[ncore:nocc,ncore:nocc,ncore:nocc,ncore:nocc] = casdm2
    for i in range(ncore):
        for j in range(ncore):
            dm2[i,i,j,j] += 4
            dm2[i,j,j,i] += -2
        dm2[i,i,ncore:nocc,ncore:nocc] = dm2[ncore:nocc,ncore:nocc,i,i] =2*casdm1
        dm2[i,ncore:nocc,ncore:nocc,i] = dm2[ncore:nocc,i,i,ncore:nocc] = -casdm1
    return dm1, dm2

# In AO representation
def make_rdm12(casscf, mo_coeff=None, ci=None):
    if ci is None: ci = casscf.ci
    if mo_coeff is None: mo_coeff = casscf.mo_coeff
    assert (not _is_uhf_mo(mo_coeff))
    nelecas = casscf.nelecas
    ncas = casscf.ncas
    ncore = casscf.ncore
    nmo = mo_coeff.shape[1]
    casdm1, casdm2 = casscf.fcisolver.make_rdm12(ci, ncas, nelecas)
    rdm1, rdm2 = _make_rdm12_on_mo(casdm1, casdm2, ncore, ncas, nmo)
    rdm1 = reduce(numpy.dot, (mo_coeff, rdm1, mo_coeff.T))
    rdm2 = numpy.dot(mo_coeff, rdm2.reshape(nmo,-1))
    rdm2 = numpy.dot(rdm2.reshape(-1,nmo), mo_coeff.T)
    rdm2 = rdm2.reshape(nmo,nmo,nmo,nmo).transpose(2,3,0,1)
    rdm2 = numpy.dot(mo_coeff, rdm2.reshape(nmo,-1))
    rdm2 = numpy.dot(rdm2.reshape(-1,nmo), mo_coeff.T)
    return rdm1, rdm2.reshape(nmo,nmo,nmo,nmo)

def get_fock(casscf, mo_coeff=None, ci=None):
    '''Generalized Fock matrix in AO representation
    '''
    if mo_coeff is None: mo_coeff = casscf.mo_coeff
    if _is_uhf_mo(mo_coeff):
        raise RuntimeError('TODO: UCAS general fock')
    else:
        return casscf.get_fock(mo_coeff, ci)

def cas_natorb(casscf, mo_coeff=None, ci=None, sort=False):
    '''Natrual orbitals in CAS space
    '''
    if mo_coeff is None: mo_coeff = casscf.mo_coeff
    if _is_uhf_mo(mo_coeff):
        raise RuntimeError('TODO: UCAS natrual orbitals')
    else:
        return casscf.cas_natorb(mo_coeff, ci, sort=sort)

def map2hf(casscf, mf_mo=None, base=BASE, tol=MAP2HF_TOL):
    '''The overlap between the CASSCF optimized orbitals and the canonical HF orbitals.
    '''
    if mf_mo is None: mf_mo = casscf._scf.mo_coeff
    s = casscf.mol.intor_symmetric('int1e_ovlp')
    s = reduce(numpy.dot, (casscf.mo_coeff.T, s, mf_mo))
    idx = numpy.argwhere(abs(s) > tol)
    for i,j in idx:
        logger.info(casscf, '<mo_coeff-mcscf|mo_coeff-hf>  %d  %d  %12.8f',
                    i+base, j+base, s[i,j])
    return idx

def spin_square(casscf, mo_coeff=None, ci=None, ovlp=None):
    '''Spin square of the UHF-CASSCF wavefunction

    Returns:
        A list of two floats.  The first is the expectation value of S^2.
        The second is the corresponding 2S+1

    Examples:

    >>> from pyscf import gto, scf, mcscf
    >>> mol = gto.M(atom='O 0 0 0; O 0 0 1', basis='sto-3g', spin=2, verbose=0)
    >>> mf = scf.UHF(mol)
    >>> res = mf.scf()
    >>> mc = mcscf.CASSCF(mf, 4, 6)
    >>> res = mc.kernel()
    >>> print('S^2 = %.7f, 2S+1 = %.7f' % mcscf.spin_square(mc))
    S^2 = 3.9831589, 2S+1 = 4.1149284
    '''
    if ci is None: ci = casscf.ci
    ncore    = casscf.ncore
    ncas     = casscf.ncas
    nelecas  = casscf.nelecas
    if isinstance(ncore, (int, numpy.integer)):
        return fci.spin_op.spin_square0(ci, ncas, nelecas)
    else:
        if mo_coeff is None: mo_coeff = casscf.mo_coeff
        if ovlp is None: ovlp = casscf._scf.get_ovlp()
        nocc = (ncore[0] + ncas, ncore[1] + ncas)
        mocas = (mo_coeff[0][:,ncore[0]:nocc[0]], mo_coeff[1][:,ncore[1]:nocc[1]])
        if isinstance(ci, (list, tuple, RANGE_TYPE)):
            sscas = numpy.array([fci.spin_op.spin_square(c, ncas, nelecas, mocas, ovlp)[0]
                                 for c in ci])
        else:
            sscas = fci.spin_op.spin_square(ci, ncas, nelecas, mocas, ovlp)[0]
        mocore = (mo_coeff[0][:,:ncore[0]], mo_coeff[1][:,:ncore[1]])
        sscore = casscf._scf.spin_square(mocore, ovlp)[0]
        logger.debug(casscf, 'S^2 of core %s  S^2 of cas %s', sscore, sscas)
        ss = sscas+sscore
        s = numpy.sqrt(ss+.25) - .5
        return ss, s*2+1

# A tag to label the derived FCI class
class StateAverageFCISolver:
    pass
class StateAverageMixFCISolver(StateAverageFCISolver):
    pass
class StateSpecificFCISolver:
    pass
# A tag to label the derived MCSCF class
class StateAverageMCSCFSolver:
    pass

def state_average(casscf, weights=(0.5,0.5), wfnsym=None):
    ''' State average over the energy.  The energy funcitonal is
    E = w1<psi1|H|psi1> + w2<psi2|H|psi2> + ...

    Note we may need change the FCI solver to

    mc.fcisolver = fci.solver(mol, False)

    before calling state_average_(mc), to mix the singlet and triplet states

    MRH, 04/08/2019: Instead of turning casscf._finalize into an instance attribute
    that points to the previous casscf object, I'm going to make a whole new child class.
    This will have the added benefit of making state_average and state_average_
    actually behave differently for the first time (until now they *both* modified the
    casscf object inplace). I'm also going to assign the weights argument as a member
    of the mc child class because an accurate second-order CASSCF algorithm for state-averaged
    calculations requires that the gradient and Hessian be computed for CI vectors of each root
    individually and then multiplied by that root's weight. The second derivatives computed
    by newton_casscf.py need to be extended to state-averaged calculations in order to be
    used as intermediates for calculations of the gradient of a single root in the context
    of the SA-CASSCF method; see: Mol. Phys. 99, 103 (2001).
    '''
    assert (abs(sum(weights)-1) < 1e-3)
    fcibase_class = casscf.fcisolver.__class__
    has_spin_square = getattr(casscf.fcisolver, 'spin_square', None)
    if wfnsym is None:
        wfnsym = casscf.fcisolver.wfnsym

    class FakeCISolver(fcibase_class, StateAverageFCISolver):
        def __init__(self, fcibase):
            self.__dict__.update (fcibase.__dict__)
            self.nroots = len(weights)
            self.weights = weights
            self.wfnsym = wfnsym
            self.e_states = [None]
            # MRH 09/09/2022: I turned the _base_class property into an
            # attribute to prevent conflict with fix_spin_ dynamic class
            self._base_class = fcibase_class
            keys = set (('weights','e_states','_base_class'))
            self._keys = self._keys.union (keys)

        def dump_flags(self, verbose=None):
            if hasattr(fcibase_class, 'dump_flags'):
                fcibase_class.dump_flags(self, verbose)
            log = logger.new_logger(self, verbose)
            log.info('State-average over %d states with weights %s',
                     len(self.weights), self.weights)
            return self

        def kernel(self, h1, h2, norb, nelec, ci0=None, **kwargs):
            if 'nroots' not in kwargs:
                kwargs['nroots'] = self.nroots

            # call fcibase_class.kernel function because the attribute orbsym
            # is available in self but undefined in fcibase object
            e, c = fcibase_class.kernel(self, h1, h2, norb, nelec, ci0=ci0,
                                        wfnsym=self.wfnsym, **kwargs)
            self.e_states = e

            log = logger.new_logger(self, kwargs.get('verbose'))
            if log.verbose >= logger.DEBUG:
                if has_spin_square:
                    ss = self.states_spin_square(c, norb, nelec)[0]
                    for i, ei in enumerate(e):
                        log.debug('state %d  E = %.15g S^2 = %.7f', i, ei, ss[i])
                else:
                    for i, ei in enumerate(e):
                        log.debug('state %d  E = %.15g', i, ei)
            return numpy.einsum('i,i->', e, self.weights), c

        def approx_kernel(self, h1, h2, norb, nelec, ci0=None, **kwargs):
            try:
                e, c = fcibase_class.approx_kernel(self, h1, h2, norb, nelec,
                                                   ci0=ci0, nroots=self.nroots,
                                                   wfnsym=self.wfnsym,
                                                   **kwargs)
            except AttributeError:
                e, c = fcibase_class.kernel(self, h1, h2, norb, nelec, ci0=ci0,
                                            nroots=self.nroots,
                                            wfnsym=self.wfnsym, **kwargs)
            return numpy.einsum('i,i->', e, self.weights), c

        def states_make_rdm1(self, ci0, norb, nelec, *args, **kwargs):
            dm1 = [fcibase_class.make_rdm1(self, c, norb, nelec, *args, **kwargs) for c in ci0]
            return dm1

        def make_rdm1(self, ci0, norb, nelec, *args, **kwargs):
            return sum ([w * dm for w, dm in zip(self.weights,
                                                 self.states_make_rdm1(ci0, norb, nelec, *args, **kwargs))])

        def states_make_rdm1s(self, ci0, norb, nelec, *args, **kwargs):
            dm1a = []
            dm1b = []
            for c in ci0:
                dm1s = fcibase_class.make_rdm1s(self, c, norb, nelec, *args, **kwargs)
                dm1a.append (dm1s[0])
                dm1b.append (dm1s[1])
            return dm1a, dm1b

        def make_rdm1s(self, ci0, norb, nelec, *args, **kwargs):
            dm1s = self.states_make_rdm1s(ci0, norb, nelec, *args, **kwargs)
            dm1s = numpy.einsum ('r,srpq->spq', self.weights, dm1s)
            return dm1s[0], dm1s[1]

        def states_make_rdm12(self, ci0, norb, nelec, *args, **kwargs):
            rdm1 = []
            rdm2 = []
            for c in ci0:
                dm1, dm2 = fcibase_class.make_rdm12(self, c, norb, nelec, *args, **kwargs)
                rdm1.append (dm1)
                rdm2.append (dm2)
            return rdm1, rdm2

        def make_rdm12(self, ci0, norb, nelec, *args, **kwargs):
            rdm1, rdm2 = self.states_make_rdm12(ci0, norb, nelec, *args, **kwargs)
            rdm1 = numpy.einsum ('r,rpq->pq', self.weights, rdm1)
            rdm2 = numpy.einsum ('r,rpqst->pqst', self.weights, rdm2)
            return rdm1, rdm2

        def states_make_rdm12s(self, ci0, norb, nelec, *args, **kwargs):
            dm1a, dm1b = [], []
            dm2aa, dm2ab, dm2bb = [], [], []
            for c in ci0:
                dm1s, dm2s = fcibase_class.make_rdm12s(self, c, norb, nelec, *args, **kwargs)
                dm1a.append(dm1s[0])
                dm1b.append(dm1s[1])
                dm2aa.append(dm2s[0])
                dm2ab.append(dm2s[1])
                dm2bb.append(dm2s[2])
            return (dm1a, dm1b), (dm2aa, dm2ab, dm2bb)

        def make_rdm12s(self, ci0, norb, nelec, *args, **kwargs):
            rdm1s, rdm2s = self.states_make_rdm12s(ci0, norb, nelec, *args, **kwargs)
            rdm1s = numpy.einsum ('r,srpq->spq', self.weights, rdm1s)
            rdm2s = numpy.einsum ('r,srpqtu->spqtu', self.weights, rdm2s)
            return rdm1s, rdm2s

        def states_trans_rdm12 (self, ci1, ci0, norb, nelec, *args, **kwargs):
            tdm1 = []
            tdm2 = []
            for c1, c0 in zip (ci1, ci0):
                dm1, dm2 = fcibase_class.trans_rdm12 (self, c1, c0, norb, nelec)
                tdm1.append (dm1)
                tdm2.append (dm2)
            return tdm1, tdm2

        def trans_rdm12 (self, ci1, ci0, norb, nelec, *args, **kwargs):
            tdm1, tdm2 = self.states_trans_rdm12 (ci1, ci0, norb, nelec, *args, **kwargs)
            tdm1 = numpy.einsum ('r,rpq->pq', self.weights, tdm1)
            tdm2 = numpy.einsum ('r,rpqst->pqst', self.weights, tdm2)
            return tdm1, tdm2

        if has_spin_square:
            def spin_square(self, ci0, norb, nelec, *args, **kwargs):
                ss, multip = self.states_spin_square(ci0, norb, nelec, *args, **kwargs)
                weights = self.weights
                return numpy.dot(ss, weights), numpy.dot(multip, weights)

            def states_spin_square(self, ci0, norb, nelec, *args, **kwargs):
                s = [fcibase_class.spin_square(self, ci0[i], norb, nelec, *args, **kwargs)
                     for i, wi in enumerate(self.weights)]
                return [x[0] for x in s], [x[1] for x in s]

    # No recursion of FakeCISolver is allowed!
    if isinstance (casscf.fcisolver, StateAverageFCISolver):
        fcisolver = casscf.fcisolver
        fcisolver.nroots = len(weights)
        fcisolver.weights = weights
    else:
        fcisolver = FakeCISolver(casscf.fcisolver)
    return _state_average_mcscf_solver(casscf, fcisolver)

def _state_average_mcscf_solver(casscf, fcisolver):
    '''A common routine for function state_average and state_average_mix to
    generate state-average MCSCF solver.
    '''
    mcscfbase_class = casscf.__class__
    if isinstance (casscf, StateAverageMCSCFSolver):
        raise TypeError('mc is not base MCSCF solver\n'
                        'state_average function cannot work with decorated '
                        'MCSCF solver %s.\nYou can restore the base MCSCF '
                        'then call state_average function, eg\n'
                        '    mc = %s.%s(mc._scf, %s, %s)\n'
                        '    mc.state_average_()\n' %
                        (mcscfbase_class, mcscfbase_class.__base__.__module__,
                         mcscfbase_class.__base__.__name__, casscf.ncas, casscf.nelecas))

    has_spin_square = getattr(fcisolver, 'spin_square', None)

    class StateAverageMCSCF(mcscfbase_class, StateAverageMCSCFSolver):
        def __init__(self, my_mc):
            self.__dict__.update (my_mc.__dict__)
            self.fcisolver = fcisolver
            keys = set (('weights', '_base_class'))
            self._keys = self._keys.union (keys)

        @property
        def _base_class (self):
            ''' for convenience; this is equal to mcscfbase_class '''
            return self.__class__.__bases__[0]

        @property
        def weights (self):
            ''' I want these to be accessible but not separable from fcisolver.weights '''
            return self.fcisolver.weights

        @weights.setter
        def weights (self, x):
            self.fcisolver.weights = x
            return self.fcisolver.weights

        @property
        def e_average(self):
            return numpy.dot(self.fcisolver.weights, self.fcisolver.e_states)

        @property
        def e_states(self):
            return self.fcisolver.e_states

        def _finalize(self):
            mcscfbase_class._finalize(self)
            # Do not overwrite self.e_tot because .e_tot needs to be used in
            # geometry optimization. self.e_states can be used to access the
            # energy of each state
            #self.e_tot = self.fcisolver.e_states
            logger.note(self, 'CASCI state-averaged energy = %.15g', self.e_average)
            logger.note(self, 'CASCI energy for each state')
            if has_spin_square:
                ss = self.fcisolver.states_spin_square(self.ci, self.ncas,
                                                       self.nelecas)[0]
                for i, ei in enumerate(self.e_states):
                    logger.note(self, '  State %d weight %g  E = %.15g S^2 = %.7f',
                                i, self.weights[i], ei, ss[i])
            else:
                for i, ei in enumerate(self.e_states):
                    logger.note(self, '  State %d weight %g  E = %.15g',
                                i, self.weights[i], ei)
            return self

        def nuc_grad_method (self, state=None):
            if callable (getattr (self, '_state_average_nuc_grad_method', None)):
                return self._state_average_nuc_grad_method (state=state)
            else: # Avoid messing up state-average CASCI
                return self._base_class.nuc_grad_method (self)

        Gradients = nuc_grad_method

    return StateAverageMCSCF(casscf)

def state_average_(casscf, weights=(0.5,0.5), wfnsym=None):
    ''' Inplace version of state_average '''
    sacasscf = state_average (casscf, weights, wfnsym)
    casscf.__class__ = sacasscf.__class__
    casscf.__dict__.update (sacasscf.__dict__)
    return casscf


def state_specific_(casscf, state=1, wfnsym=None):
    '''For excited state

    Kwargs:
        state : int
        0 for ground state; 1 for first excited state.
    '''
    fcibase_class = casscf.fcisolver.__class__
    if fcibase_class.__name__ == 'FakeCISolver':
        raise TypeError('mc.fcisolver is not base FCI solver\n'
                        'state_specific function cannot work with decorated '
                        'fcisolver %s.\nYou can restore the base fcisolver '
                        'then call state_specific function, eg\n'
                        '    mc.fcisolver = %s.%s(mc.mol)\n'
                        '    mc.state_specific_()\n' %
                        (casscf.fcisolver, fcibase_class.__base__.__module__,
                         fcibase_class.__base__.__name__))
    if wfnsym is None:
        wfnsym = casscf.fcisolver.wfnsym

    class FakeCISolver(fcibase_class, StateSpecificFCISolver):
        def __init__(self):
            self.nroots = state+1
            self._civec = None
            self.wfnsym = wfnsym

        def kernel(self, h1, h2, norb, nelec, ci0=None, **kwargs):
            if self._civec is not None:
                ci0 = self._civec
            e, c = fcibase_class.kernel(self, h1, h2, norb, nelec, ci0=ci0,
                                        nroots=self.nroots, wfnsym=self.wfnsym,
                                        **kwargs)
            if state == 0:
                e = [e]
                c = [c]
            self._civec = c
            log = logger.new_logger(self, kwargs.get('verbose'))
            if log.verbose >= logger.DEBUG:
                if getattr(fcibase_class, 'spin_square', None):
                    ss = fcibase_class.spin_square(self, c[state], norb, nelec)
                    log.debug('state %d  E = %.15g S^2 = %.7f',
                              state, e[state], ss[0])
                else:
                    log.debug('state %d  E = %.15g', state, e[state])
            return e[state], c[state]

        def approx_kernel(self, h1, h2, norb, nelec, ci0=None, **kwargs):
            if self._civec is not None:
                ci0 = self._civec
            try:
                e, c = fcibase_class.approx_kernel(self, h1, h2, norb, nelec,
                                                   ci0=ci0, nroots=self.nroots,
                                                   wfnsym=self.wfnsym,
                                                   **kwargs)
            except AttributeError:
                e, c = fcibase_class.kernel(self, h1, h2, norb, nelec, ci0=ci0,
                                            nroots=self.nroots,
                                            wfnsym=self.wfnsym, **kwargs)
            if state == 0:
                self._civec = [c]
                return e, c
            else:
                self._civec = c
                return e[state], c[state]

    fcisolver = FakeCISolver()
    fcisolver.__dict__.update(casscf.fcisolver.__dict__)
    fcisolver.nroots = state + 1
    if wfnsym is not None:
        fcisolver.wfnsym = wfnsym
    casscf.fcisolver = fcisolver
    return casscf
state_specific = state_specific_

class StateAverageMixFCISolver_solver_args:
    def __init__(self, data):
        self._data = data
    def __getitem__(self, key): # Handle data = None
        try:
            return self._data[key]
        except TypeError as e:
            assert (self._data is None), e
            return None
class StateAverageMixFCISolver_state_args (StateAverageMixFCISolver_solver_args):
    pass

def state_average_mix(casscf, fcisolvers, weights=(0.5,0.5)):
    '''State-average CASSCF over multiple FCI solvers.
    '''
    fcibase_class = fcisolvers[0].__class__
    nroots = sum(solver.nroots for solver in fcisolvers)
    assert (nroots == len(weights))
    has_spin_square = all(getattr(solver, 'spin_square', None)
                          for solver in fcisolvers)
    has_large_ci = all(getattr(solver, 'large_ci', None)
                       for solver in fcisolvers)
    has_transform_ci = all(getattr(solver, 'transform_ci_for_orbital_rotation', None)
                           for solver in fcisolvers)
    _solver_args = StateAverageMixFCISolver_solver_args
    _state_args = StateAverageMixFCISolver_state_args
    class FakeCISolver(fcibase_class, StateAverageMixFCISolver):
        def __init__(self, mol):
            fcibase_class.__init__(self, mol)
            self.nroots = len(weights)
            self.weights = weights
            self.e_states = [None]
            self.fcisolvers = fcisolvers
            keys = set (('weights','e_states','_base_class','fcisolvers'))
            self._keys = self._keys.union (keys)

        @property
        def _base_class (self):
            ''' for convenience; this is equal to mcscfbase_class '''
            return self.__class__.__bases__[0]

        # MRH 06/24/2020: I need these functions in newton_casscf!
        # TODO: handle things like linkstr somehow (variables that
        # have to be different for different solvers or ci vecs)
        def _loop_solver(self, *args, **kwargs):
            p0 = 0
            for ix, solver in enumerate (self.fcisolvers):
                my_args = []
                for arg in args:
                    if isinstance (arg, _state_args):
                        my_arg = arg[p0:p0+solver.nroots]
                        if solver.nroots == 1 and my_arg is not None: my_arg = my_arg[0]
                        my_args.append (my_arg)
                    elif isinstance (arg, _solver_args):
                        my_args.append (arg[ix])
                    else:
                        my_args.append (arg)
                my_kwargs = {}
                for key, item in kwargs.items ():
                    if isinstance (item, _state_args):
                        my_arg = item[p0:p0+solver.nroots]
                        if solver.nroots == 1 and my_arg is not None: my_arg = my_arg[0]
                        my_kwargs[key] = my_arg
                    elif isinstance (item, _solver_args):
                        my_kwargs[key] = item[ix]
                    else:
                        my_kwargs[key] = item
                yield solver, my_args, my_kwargs
                p0 += solver.nroots

        def _loop_civecs(self, *args, **kwargs):
            p0 = 0
            for i, solver in enumerate (self.fcisolvers):
                for j in range(p0, p0+solver.nroots):
                    my_args = []
                    for arg in args:
                        if isinstance (arg, _state_args):
                            my_args.append (arg[j])
                        elif isinstance (arg, _solver_args):
                            my_args.append (arg[i])
                        else:
                            my_args.append (arg)
                    my_kwargs = {}
                    for key, item in kwargs.items ():
                        if isinstance (item, _state_args):
                            my_kwargs[key] = item[j]
                        elif isinstance (item, _solver_args):
                            my_kwargs[key] = item[i]
                        else:
                            my_kwargs[key] = item
                    yield solver, my_args, my_kwargs
                p0 += solver.nroots

        def _get_nelec(self, solver, nelec):
            # FCISolver does not need this function. Some external solver may not
            # have the function to handle nelec and spin
            # MRH 06/24/2020: Yes, FCISolver DOES need this function!
            if solver.spin is not None:
                nelec = numpy.sum(nelec)
                nelec = (nelec+solver.spin)//2, (nelec-solver.spin)//2
            return nelec

        def _collect(self, fname, *args, **kwargs):
            for solver, args, kwargs in self._loop_civecs(*args, **kwargs):
                fn = getattr(solver, fname)
                yield fn(*args, **kwargs)

        def kernel(self, h1, h2, norb, nelec, ci0=None, verbose=0, **kwargs):
            # Note self.orbsym is initialized lazily in mc1step_symm.kernel function
            log = logger.new_logger(self, verbose)
            es = []
            cs = []
            for solver, my_args, my_kwargs in self._loop_solver(_state_args (ci0)):
                c0 = my_args[0]
                e, c = solver.kernel(h1, h2, norb, self._get_nelec(solver, nelec), ci0=c0,
                                     orbsym=self.orbsym, verbose=log, **kwargs)
                if solver.nroots == 1:
                    es.append(e)
                    cs.append(c)
                else:
                    es.extend(e)
                    cs.extend(c)
            self.e_states = es
            self.converged = numpy.all(getattr(sol, 'converged', True)
                                       for sol in fcisolvers)

            if log.verbose >= logger.DEBUG:
                if has_spin_square:
                    ss = self.states_spin_square(cs, norb, nelec)[0]
                    for i, ei in enumerate(es):
                        log.debug('state %d  E = %.15g S^2 = %.7f', i, ei, ss[i])
                else:
                    for i, ei in enumerate(es):
                        log.debug('state %d  E = %.15g', i, ei)
            return numpy.einsum('i,i', numpy.array(es), weights), cs

        def approx_kernel(self, h1, h2, norb, nelec, ci0=None, **kwargs):
            es = []
            cs = []
            for ix, (solver, my_args, my_kwargs) in enumerate (self._loop_solver(_state_args (ci0))):
                c0 = my_args[0]
                try:
                    e, c = solver.approx_kernel(h1, h2, norb, self._get_nelec(solver, nelec), ci0=c0,
                                                orbsym=self.orbsym, **kwargs)
                except AttributeError:
                    e, c = solver.kernel(h1, h2, norb, self._get_nelec(solver, nelec), ci0=c0,
                                         orbsym=self.orbsym, **kwargs)
                if solver.nroots == 1:
                    es.append(e)
                    cs.append(c)
                else:
                    es.extend(e)
                    cs.extend(c)
            return numpy.einsum('i,i->', es, weights), cs

        def states_make_rdm1 (self, ci0, norb, nelec, link_index=None, **kwargs):
            ci0 = _state_args (ci0)
            link_index = _solver_args (link_index)
            nelec = _solver_args ([self._get_nelec (solver, nelec) for solver in self.fcisolvers])
            return [dm for dm in self._collect ('make_rdm1', ci0, norb, nelec, link_index=link_index, **kwargs)]

        def make_rdm1(self, ci0, norb, nelec, link_index=None, **kwargs):
            dm1 = self.states_make_rdm1 (ci0, norb, nelec, link_index=link_index, **kwargs)
            return numpy.einsum ('r,rpq->pq', self.weights, dm1)

        def states_make_rdm1s (self, ci0, norb, nelec, link_index=None, **kwargs):
            ci0 = _state_args (ci0)
            link_index = _solver_args (link_index)
            nelec = _solver_args ([self._get_nelec (solver, nelec) for solver in self.fcisolvers])
            dm1a = []
            dm1b = []
            for dm1s in self._collect ('make_rdm1s', ci0, norb, nelec, link_index=link_index, **kwargs):
                dm1a.append (dm1s[0])
                dm1b.append (dm1s[1])
            return dm1a, dm1b

        def make_rdm1s(self, ci0, norb, nelec, link_index=None, **kwargs):
            dm1a, dm1b = self.states_make_rdm1s (ci0, norb, nelec, link_index=link_index, **kwargs)
            dm1s = numpy.einsum ('r,srpq->spq', self.weights, [dm1a, dm1b])
            return dm1s[0], dm1s[1]

        def states_make_rdm12 (self, ci0, norb, nelec, link_index=None, **kwargs):
            ci0 = _state_args (ci0)
            link_index = _solver_args (link_index)
            nelec = _solver_args ([self._get_nelec (solver, nelec) for solver in self.fcisolvers])
            rdm1 = []
            rdm2 = []
            for dm1, dm2 in self._collect ('make_rdm12', ci0, norb, nelec, link_index=link_index, **kwargs):
                rdm1.append (dm1)
                rdm2.append (dm2)
            return rdm1, rdm2

        def make_rdm12(self, ci0, norb, nelec, link_index=None, **kwargs):
            rdm1, rdm2 = self.states_make_rdm12 (ci0, norb, nelec, link_index=link_index, **kwargs)
            rdm1 = numpy.einsum ('r,rpq->pq', self.weights, rdm1)
            rdm2 = numpy.einsum ('r,rpqst->pqst', self.weights, rdm2)
            return rdm1, rdm2

        def states_make_rdm12s(self, ci0, norb, nelec, link_index=None, **kwargs):
            ci0 = _state_args (ci0)
            link_index = _solver_args (link_index)
            nelec = _solver_args ([self._get_nelec (solver, nelec) for solver in self.fcisolvers])
            dm1a, dm1b = [], []
            dm2aa, dm2ab, dm2bb = [], [], []
            for dm1s, dm2s in self._collect ('make_rdm12s', ci0, norb, nelec, link_index=link_index, **kwargs):
                dm1a.append(dm1s[0])
                dm1b.append(dm1s[1])
                dm2aa.append(dm2s[0])
                dm2ab.append(dm2s[1])
                dm2bb.append(dm2s[2])
            return (dm1a, dm1b), (dm2aa, dm2ab, dm2bb)

        def make_rdm12s(self, ci0, norb, nelec, link_index=None, **kwargs):
            rdm1s, rdm2s = self.states_make_rdm12s(ci0, norb, nelec, link_index=link_index, **kwargs)
            rdm1s = numpy.einsum ('r,srpq->spq', self.weights, rdm1s)
            rdm2s = numpy.einsum ('r,srpqtu->spqtu', self.weights, rdm2s)
            return rdm1s, rdm2s

        # TODO: linkstr support
        def states_trans_rdm12 (self, ci1, ci0, norb, nelec, link_index=None, **kwargs):
            ci1 = _state_args (ci1)
            ci0 = _state_args (ci0)
            link_index = _solver_args (link_index)
            nelec = _solver_args ([self._get_nelec (solver, nelec) for solver in self.fcisolvers])
            tdm1 = []
            tdm2 = []
            for dm1, dm2 in self._collect ('trans_rdm12', ci1, ci0, norb, nelec, link_index=link_index, **kwargs):
                tdm1.append (dm1)
                tdm2.append (dm2)
            return tdm1, tdm2

        def trans_rdm12 (self, ci1, ci0, norb, nelec, link_index=None, **kwargs):
            tdm1, tdm2 = self.states_trans_rdm12 (ci1, ci0, norb, nelec, link_index=link_index, **kwargs)
            tdm1 = numpy.einsum ('r,rpq->pq', self.weights, tdm1)
            tdm2 = numpy.einsum ('r,rpqst->pqst', self.weights, tdm2)
            return tdm1, tdm2

        if has_spin_square:
            def spin_square(self, ci0, norb, nelec, *args, **kwargs):
                ss, multip = self.states_spin_square(ci0, norb, nelec, *args, **kwargs)
                weights = self.weights
                return numpy.dot(ss, weights), numpy.dot(multip, weights)

            def states_spin_square(self, ci0, norb, nelec, *args, **kwargs):
                ci0 = _state_args (ci0)
                nelec = _solver_args ([self._get_nelec (solver, nelec) for solver in self.fcisolvers])
                res = list(self._collect('spin_square', ci0, norb, nelec, *args, **kwargs))
                ss = [x[0] for x in res]
                multip = [x[1] for x in res]
                return ss, multip
        else:
            spin_square = None

        large_ci = None
        if has_large_ci:
            def states_large_ci(self, fcivec, norb, nelec, *args, **kwargs):
                fcivec = _state_args (fcivec)
                nelec = _solver_args ([self._get_nelec (solver, nelec) for solver in self.fcisolvers])
                return list(self._collect('large_ci', fcivec, norb, nelec, *args, **kwargs))

        transform_ci_for_orbital_rotation = None
        if has_transform_ci:
            def states_transform_ci_for_orbital_rotation(self, fcivec, norb, nelec,
                                                         *args, **kwargs):
                fcivec = _state_args (fcivec)
                nelec = _solver_args ([self._get_nelec (solver, nelec) for solver in self.fcisolvers])
                return list(self._collect('transform_ci_for_orbital_rotation',
                                          fcivec, norb, nelec, *args, **kwargs))

    fcisolver = FakeCISolver(casscf.mol)
    fcisolver.__dict__.update(casscf.fcisolver.__dict__)
    fcisolver.nroots = nroots
    mc = _state_average_mcscf_solver(casscf, fcisolver)
    return mc

def state_average_mix_(casscf, fcisolvers, weights=(0.5,0.5)):
    ''' Inplace version of state_average '''
    sacasscf = state_average_mix(casscf, fcisolvers, weights)
    casscf.__class__ = sacasscf.__class__
    casscf.__dict__.update(sacasscf.__dict__)
    return casscf


del (BASE, MAP2HF_TOL)


if __name__ == '__main__':
    from pyscf import mcscf
    from pyscf.tools import ring

    mol = gto.M(verbose=0,
                output=None,
                atom=[['H', c] for c in ring.make(6, 1.2)],
                basis='6-31g')

    m = scf.RHF(mol)
    ehf = m.scf()

    mc = mcscf.CASSCF(m, 6, 6)
    mc.verbose = 4
    emc, e_ci, fcivec, mo, mo_energy = mc.mc1step()
    print(ehf, emc, emc-ehf)
    print(emc - -3.272089958)

    rdm1 = make_rdm1(mc, mo, fcivec)
    rdm1, rdm2 = make_rdm12(mc, mo, fcivec)
    print(rdm1)

    mo1 = cas_natorb(mc)[0]
    numpy.set_printoptions(2)
    print(reduce(numpy.dot, (mo1[:,:6].T, mol.intor('int1e_ovlp_sph'),
                             mo[:,:6])))

# state average
    mol.atom = [
        ['O', ( 0., 0.    , 0.   )],
        ['H', ( 0., -0.757, 0.587)],
        ['H', ( 0., 0.757 , 0.587)],]
    mol.basis = '6-31g'
    mol.symmetry = 1
    mol.build()

    m = scf.RHF(mol)
    ehf = m.scf()


    mc = mcscf.CASSCF(m, 4, 4)
    mc.verbose = 4
    mc.fcisolver = fci.solver(mol, False) # to mix the singlet and triplet
    mc = state_average_(mc, (.64,.36))
    emc, e_ci, fcivec, mo, mo_energy = mc.mc1step()[:5]
    mc = mcscf.CASCI(m, 4, 4)
    emc = mc.casci(mo)[0]
    print(ehf, emc, emc-ehf)
    print(emc - -76.003352190262532)

    mc = mcscf.CASSCF(m, 4, 4)
    mc.verbose = 4
    mc = state_average_(mc, (.64,.36))
    emc, e_ci, fcivec, mo, mo_energy = mc.mc1step()[:5]
    mc = mcscf.CASCI(m, 4, 4)
    emc = mc.casci(mo)[0]
    print(ehf, emc, emc-ehf)
    print(emc - -75.982520334896776)


    mc = mcscf.CASSCF(m, 4, 4)
    mc.verbose = 4
    mc = state_specific_(mc, 2)
    emc = mc.kernel()[0]




#!/usr/bin/env python
# Copyright 2014-2019 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
Analytical nuclear gradients
============================

Simple usage::

    >>> from pyscf import gto, scf, grad
    >>> mol = gto.M(atom='N 0 0 0; N 0 0 1', basis='ccpvdz')
    >>> mf = scf.RHF(mol).run()
    >>> grad.RHF(mf).kernel()
'''

from . import rhf
from . import dhf
from . import uhf
from . import rohf
from .rhf import Gradients as RHF
from .dhf import Gradients as DHF
from .uhf import Gradients as UHF
from .rohf import Gradients as ROHF

grad_nuc = rhf.grad_nuc

try:
    from . import casci
    from . import casscf
    from . import sacasscf
    from . import ccsd
    #from . import ccsd_t
    from . import cisd
    from . import mp2
    from . import rks
    from . import roks
    from . import tdrhf
    from . import tdrks
    from . import tduhf
    from . import tduks
    from . import uccsd
    #from . import uccsd_t
    from . import ucisd
    from . import uks
    from . import ump2

    from .rks import Gradients as RKS
    from .uks import Gradients as UKS
    from .roks import Gradients as ROKS

except (ImportError, OSError):
    pass



# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Oliver Backhouse <olbackhouse@gmail.com>
#         George Booth <george.booth@kcl.ac.uk>
#

'''
Functions for tuning the chemical potential.
'''

import numpy as np
from scipy import optimize
from pyscf import lib
from pyscf.lib import logger
from pyscf import __config__


def _objective(x, se, fock, nelec, occupancy=2, buf=None):
    ''' Objective function for the minimization
    '''

    w, v = se.eig(fock, chempot=x, out=buf)

    chempot, error = binsearch_chempot((w,v), se.nphys, nelec,
                                       occupancy=occupancy)

    return error**2


def _gradient(x, se, fock, nelec, occupancy=2, buf=None):
    ''' Gradient function for the minimization
    '''

    w, v = se.eig(fock, chempot=x, out=buf)

    chempot, error = binsearch_chempot((w,v), se.nphys, nelec,
                                       occupancy=occupancy)

    nocc = np.sum(w < chempot)
    nphys = se.nphys

    h1 = -np.dot(v[nphys:,nocc:].conj().T, v[nphys:,:nocc])
    zai = -h1 / lib.direct_sum('i,a->ai', w[:nocc], -w[nocc:])

    c_occ = np.dot(v[:nphys,nocc:], zai)
    d_rdm1 = np.dot(v[:nphys,:nocc], c_occ.conj().T) * 4

    ne = np.trace(d_rdm1).real
    d = occupancy * error * ne

    return error**2, d


def binsearch_chempot(fock, nphys, nelec, occupancy=2):
    ''' Finds a chemical potential which best agrees with the number
        of physical electrons and abides by the Aufbau principal via
        a binary search.

    Args:
        fock : 2D array or tuple of arrays
            Fock matrix to diagonalise, may be the physical Fock matrix
            or extended Fock matrix. Can also be the output of
            :func:`np.linalg.eigh` for this matrix, i.e. a tuple of the
            eigenvalues and eigenvectors.
        nphys : int
            Number of physical degrees of freedom
        nelec : int
            Number of physical electrons

    Kwargs:
        occupancy : int
            Occupancy of the states, i.e. 2 for RHF and 1 for UHF.
            Default 2.

    Returns:
        chemical potential, and the error in the number of electrons
    '''

    if isinstance(fock, tuple):
        w, v = fock
    else:
        w, v = np.linalg.eigh(fock)

    nmo = v.shape[-1]
    sum0 = sum1 = 0.0

    for i in range(nmo):
        n = occupancy * np.dot(v[:nphys,i].conj().T, v[:nphys,i]).real
        sum0, sum1 = sum1, sum1 + n

        if i > 0:
            if sum0 <= nelec and nelec <= sum1:
                break

    if abs(sum0 - nelec) < abs(sum1 - nelec):
        homo = i-1
        error = nelec - sum0
    else:
        homo = i
        error = nelec - sum1

    lumo = homo+1
    chempot = 0.5 * (w[homo] + w[lumo])

    return chempot, error


def minimize_chempot(se, fock, nelec, occupancy=2, x0=0.0, tol=1e-6, maxiter=200, jac=True):
    ''' Finds a set of auxiliary energies and chemical potential on
        the physical space which best satisfy the number of electrons.

    Args:
        se : AuxiliarySpace
            Auxiliary space
        fock : 2D array
            phys : 2D array
                Physical space (1p + 1h), typically the Fock matrix
        nelec : int
            Number of physical electrons

    Kwargs:
        occupancy : int
            Occupancy of the states, i.e. 2 for RHF and 1 for UHF.
            Default 2.
        x0 : float
            Initial guess for :attr:`chempot`. Default 0.0
        tol : float
            Convergence threshold (units are the same as :attr:`nelec`).
            Default 1e-6.
        maxiter : int
            Maximum number of iterations. Default 200.
        jac : bool
            If True, use gradient. Default True.

    Returns:
        AuxiliarySpace object with altered :attr:`energy` and
        :attr:`chempot`, and the SciPy :attr:`OptimizeResult` object.
    '''

    tol = tol**2  # we minimize the squared error
    dtype = np.result_type(se.energy.dtype, se.coupling.dtype, fock.dtype)
    buf = np.zeros((se.nphys+se.naux, se.nphys+se.naux), dtype=dtype)
    fargs = (se, fock, nelec, occupancy, buf)

    options = dict(maxiter=maxiter, ftol=tol, xtol=tol, gtol=tol)
    kwargs = dict(x0=x0, method='TNC', jac=jac, options=options)
    fun = _objective if not jac else _gradient

    opt = optimize.minimize(fun, args=fargs, **kwargs)

    se.energy -= opt.x
    se.chempot = binsearch_chempot(se.eig(fock), se.nphys, nelec,
                                   occupancy=occupancy)[0]

    return se, opt



#!/usr/bin/env python
# Copyright 2014-2021 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Peter Pinski <peter.pinski@quantumsimulations.de>
#

'''
Localized molecular orbitals via Cholesky factorization.
The orbitals are usually less well localized than with Boys, Pipek-Mezey, etc.
On the other hand, the procedure is non-iterative and the result unique,
except for degeneracies.

F. Aquilante, T.B. Pedersen, J. Chem. Phys. 125, 174101 (2006)
https://doi.org/10.1063/1.2360264
'''


import numpy as np
from pyscf.lib.scipy_helper import pivoted_cholesky


def cholesky_mos(mo_coeff):
    '''
    Calculates localized orbitals through a pivoted Cholesky factorization
    of the density matrix.

    Args:
        mo_coeff: block of MO coefficients to be localized

    Returns:
        the localized MOs
    '''
    assert (mo_coeff.ndim == 2)
    nao, nmo = mo_coeff.shape

    # Factorization of a density matrix-like quantity.
    D = np.dot(mo_coeff, mo_coeff.T)
    L, piv, rank = pivoted_cholesky(D, lower=True)
    if rank < nmo:
        raise RuntimeError('rank of matrix lower than the number of orbitals')

    # Permute L back to the original order of the AOs.
    # Superfluous columns are cropped out.
    P = np.zeros((nao, nao))
    P[piv, np.arange(nao)] = 1
    mo_loc = np.dot(P, L[:, :nmo])

    return mo_loc


if __name__ == "__main__":

    import numpy
    from pyscf.gto import Mole
    from pyscf.scf import RHF
    from pyscf.tools.mo_mapping import mo_comps

    mol = Mole()
    mol.atom = '''
    C        0.681068338      0.605116159      0.307300799
    C       -0.733665805      0.654940451     -0.299036438
    C       -1.523996730     -0.592207689      0.138683275
    H        0.609941801      0.564304456      1.384183068
    H        1.228991034      1.489024155      0.015946420
    H       -1.242251083      1.542928348      0.046243898
    H       -0.662968178      0.676527364     -1.376503770
    H       -0.838473936     -1.344174292      0.500629028
    H       -2.075136399     -0.983173387     -0.703807608
    H       -2.212637905     -0.323898759      0.926200671
    O        1.368219958     -0.565620846     -0.173113101
    H        2.250134219     -0.596689848      0.204857736
    '''
    mol.basis = 'STO-3G'
    mol.build()

    mf = RHF(mol)
    mf.kernel()

    nocc = numpy.count_nonzero(mf.mo_occ > 0)
    mo_loc = cholesky_mos(mf.mo_coeff[:, :nocc])
    print('LMO    Largest coefficients')
    numpy.set_printoptions(precision=3, suppress=True, sign=' ')
    for i in range(nocc):
        li = numpy.argsort(abs(mo_loc[:, i]))
        print('{0:3d}    {1}'. format(i, mo_loc[li[:-6:-1], i]))



#  Author: Artem Pulkin
"""
This and other `_slow` modules implement the time-dependent Hartree-Fock procedure. The primary performance drawback is
that, unlike other 'fast' routines with an implicit construction of the eigenvalue problem, these modules construct
TDHF matrices explicitly via an AO-MO transformation, i.e. with a O(N^5) complexity scaling. As a result, regular
`numpy.linalg.eig` can be used to retrieve TDHF roots in a reliable fashion without any issues related to the Davidson
procedure. Several variants of TDHF are available:

 * `pyscf.tdscf.rhf_slow`: the molecular implementation;
 * `pyscf.pbc.tdscf.rhf_slow`: PBC (periodic boundary condition) implementation for RHF objects of `pyscf.pbc.scf`
   modules;
 * (this module) `pyscf.pbc.tdscf.krhf_slow_supercell`: PBC implementation for KRHF objects of `pyscf.pbc.scf` modules.
   Works with an arbitrary number of k-points but has a overhead due to an effective construction of a supercell.
 * `pyscf.pbc.tdscf.krhf_slow_gamma`: A Gamma-point calculation resembling the original `pyscf.pbc.tdscf.krhf`
   module. Despite its name, it accepts KRHF objects with an arbitrary number of k-points but finds only few TDHF roots
   corresponding to collective oscillations without momentum transfer;
 * `pyscf.pbc.tdscf.krhf_slow`: PBC implementation for KRHF objects of `pyscf.pbc.scf` modules. Works with
   an arbitrary number of k-points and employs k-point conservation (diagonalizes matrix blocks separately).
"""

from pyscf.tdscf.common_slow import TDERIMatrixBlocks, PeriodicMFMixin
from pyscf.tdscf import rhf_slow

from pyscf.pbc.lib.kpts_helper import loop_kkk
from pyscf.lib import logger

import numpy
import scipy
from itertools import product


# Convention for these modules:
# * PhysERI, PhysERI4, PhysERI8 are 2-electron integral routines computed directly (for debug purposes), with a 4-fold
#   symmetry and with an 8-fold symmetry
# * vector_to_amplitudes reshapes and normalizes the solution
# * TDRHF provides a container


class PhysERI(PeriodicMFMixin, TDERIMatrixBlocks):

    def __init__(self, model, frozen=None):
        """
        The TDHF ERI implementation performing a full transformation of integrals to Bloch functions. No symmetries are
        employed in this class.

        Args:
            model (KRHF): the base model;
            frozen (int, Iterable): the number of frozen valence orbitals or the list of frozen orbitals for all
            k-points or multiple lists of frozen orbitals for each k-point;
        """
        TDERIMatrixBlocks.__init__(self)
        PeriodicMFMixin.__init__(self, model, frozen=frozen)
        # Phys representation
        self.__full_eri_k__ = {}
        for k in loop_kkk(len(model.kpts)):
            k = k + (self.kconserv[k],)
            self.__full_eri_k__[k] = self.ao2mo_k(tuple(self.mo_coeff[j] for j in k), k)

    def ao2mo_k(self, coeff, k):
        """
        Phys ERI in MO basis.
        Args:
            coeff (Iterable): MO orbitals;
            k (Iterable): the 4 k-points MOs correspond to;

        Returns:
            ERI in MO basis.
        """
        coeff = (coeff[0], coeff[2], coeff[1], coeff[3])
        k = (k[0], k[2], k[1], k[3])
        result = self.model.with_df.ao2mo(coeff, tuple(self.model.kpts[i] for i in k), compact=False)
        return result.reshape(
            tuple(i.shape[1] for i in coeff)
        ).swapaxes(1, 2)

    def __get_mo_energies__(self, k1, k2):
        """This routine collects occupied and virtual MO energies."""
        return self.mo_energy[k1][:self.nocc[k1]], self.mo_energy[k2][self.nocc[k2]:]

    def tdhf_diag_k(self, k1, k2):
        """
        Retrieves the diagonal block.
        Args:
            k1 (int): first k-index (row);
            k2 (int): second k-index (column);

        Returns:
            The diagonal block.
        """
        # Everything is already implemented in molecular code
        return super(PhysERI, self).tdhf_diag(k1, k2)

    def tdhf_diag(self, pairs=None):
        """
        Retrieves the merged diagonal block with specified or all possible k-index pairs.
        Args:
            pairs (Iterable): pairs of k-points to assmble;

        Returns:
            The diagonal block.
        """
        if pairs is None:
            pairs = product(range(len(self.model.kpts)), range(len(self.model.kpts)))
        result = []
        for k1, k2 in pairs:
            result.append(self.tdhf_diag_k(k1, k2))
        return scipy.linalg.block_diag(*result)

    def __calc_block__(self, item, k):
        if k in self.__full_eri_k__:
            slc = tuple(slice(self.nocc[_k]) if i == 'o' else slice(self.nocc[_k], None) for i, _k in zip(item, k))
            return self.__full_eri_k__[k][slc]
        else:
            return numpy.zeros(tuple(
                self.nocc[_k] if i == 'o' else self.nmo[_k] - self.nocc[_k]
                for i, _k in zip(item, k)
            ))

    def eri_mknj_k(self, item, k):
        """
        Retrieves ERI block using 'mknj' notation.
        Args:
            item (str): a 4-character string of 'mknj' letters;
            k (Iterable): k indexes;

        Returns:
            The corresponding block of ERI (phys notation).
        """
        # Everything is already implemented in molecular code
        return super(PhysERI, self).eri_mknj(item, k)

    def eri_mknj(self, item, pairs_row=None, pairs_column=None):
        """
        Retrieves the merged ERI block using 'mknj' notation with all k-indexes.
        Args:
            item (str): a 4-character string of 'mknj' letters;
            pairs_row (Iterable): iterator for pairs of row k-points (first index in the output matrix);
            pairs_column (Iterable): iterator for pairs of column k-points (second index in the output matrix);

        Returns:
            The corresponding block of ERI (phys notation).
        """
        if pairs_row is None:
            pairs_row = product(range(len(self.model.kpts)), range(len(self.model.kpts)))
        if pairs_column is None:
            pairs_column = product(range(len(self.model.kpts)), range(len(self.model.kpts)))
        # Second index has to support re-iterations
        pairs_column = tuple(pairs_column)
        result = []
        for k1, k2 in pairs_row:
            result.append([])
            for k3, k4 in pairs_column:
                result[-1].append(self.eri_mknj_k(item, (k1, k2, k3, k4)))

        r = numpy.block(result)
        return r / len(self.model.kpts)


class PhysERI4(PhysERI):
    symmetries = [
        ((0, 1, 2, 3), False),
        ((1, 0, 3, 2), False),
        ((2, 3, 0, 1), True),
        ((3, 2, 1, 0), True),
    ]

    def __init__(self, model, frozen=None):
        """
        The TDHF ERI implementation performing a partial transformation of integrals to Bloch functions. A 4-fold
        symmetry of complex-valued wavefunctions is employed in this class.

        Args:
            model (KRHF): the base model;
            frozen (int, Iterable): the number of frozen valence orbitals or the list of frozen orbitals for all
            k-points or multiple lists of frozen orbitals for each k-point;
        """
        TDERIMatrixBlocks.__init__(self)
        PeriodicMFMixin.__init__(self, model, frozen=frozen)

    def __calc_block__(self, item, k):
        if self.kconserv[k[:3]] == k[3]:
            logger.info(self.model, "Computing {} {} ...".format(''.join(item), repr(k)))
            return self.ao2mo_k(tuple(
                self.mo_coeff[_k][:, :self.nocc[_k]] if i == "o" else self.mo_coeff[_k][:, self.nocc[_k]:]
                for i, _k in zip(item, k)
            ), k)
        else:
            return numpy.zeros(tuple(
                self.nocc[_k] if i == 'o' else self.nmo[_k] - self.nocc[_k]
                for i, _k in zip(item, k)
            ))


class PhysERI8(PhysERI4):
    symmetries = [
        ((0, 1, 2, 3), False),
        ((1, 0, 3, 2), False),
        ((2, 3, 0, 1), False),
        ((3, 2, 1, 0), False),

        ((2, 1, 0, 3), False),
        ((3, 0, 1, 2), False),
        ((0, 3, 2, 1), False),
        ((1, 2, 3, 0), False),
    ]

    def __init__(self, model, frozen=None):
        """
        The TDHF ERI implementation performing a partial transformation of integrals to Bloch functions. An 8-fold
        symmetry of real-valued wavefunctions is employed in this class.

        Args:
            model (KRHF): the base model;
            frozen (int, Iterable): the number of frozen valence orbitals or the list of frozen orbitals for all
            k-points or multiple lists of frozen orbitals for each k-point;
        """
        super(PhysERI8, self).__init__(model, frozen=frozen)


def vector_to_amplitudes(vectors, nocc, nmo):
    """
    Transforms (reshapes) and normalizes vectors into amplitudes.
    Args:
        vectors (numpy.ndarray): raw eigenvectors to transform;
        nocc (tuple): numbers of occupied orbitals;
        nmo (tuple): the total numbers of AOs per k-point;

    Returns:
        Amplitudes with the following shape: (# of roots, 2 (x or y), # of kpts, # of kpts, # of occupied orbitals,
        # of virtual orbitals).
    """
    if not all(i == nocc[0] for i in nocc):
        raise NotImplementedError("Varying occupation numbers are not implemented yet")
    nk = len(nocc)
    nocc = nocc[0]
    if not all(i == nmo[0] for i in nmo):
        raise NotImplementedError("Varying AO spaces are not implemented yet")
    nmo = nmo[0]
    vectors = numpy.asanyarray(vectors)
    vectors = vectors.reshape(2, nk, nk, nocc, nmo-nocc, vectors.shape[1])
    norm = (abs(vectors) ** 2).sum(axis=(1, 2, 3, 4))
    norm = 2 * (norm[0] - norm[1])
    vectors /= norm ** .5
    return vectors.transpose(5, 0, 1, 2, 3, 4)


class TDRHF(rhf_slow.TDRHF):
    eri4 = PhysERI4
    eri8 = PhysERI8
    v2a = staticmethod(vector_to_amplitudes)



#!/usr/bin/env python
# Copyright 2014-2018 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#
# Ref:
# Chem Phys Lett, 256, 454
# J. Mol. Struct. THEOCHEM, 914, 3
#

from pyscf import lib
from pyscf.tdscf import rhf


class TDA(rhf.TDA):
    def gen_vind(self, mf):
        # gen_vind calls get_jk functions to compute the contraction between
        # two-electron integrals and X,Y amplitudes. There are two choices for
        # the treatment of exxdiv.
        # 1. Removing exxdiv corrections in both orbital energies (in hdiag) and
        #   get_jk functions (in vind function). This treatment can make TDA the
        #   same to CIS method in which exxdiv was completely excluded when
        #   constructing Hamiltonians.
        # 2. Excluding exxdiv corrections from get_jk only. Keep its correction
        #   to orbital energy. This treatment can make the TDDFT excitation
        #   energies closed to the relevant DFT orbital energy gaps.
        # DFT orbital energy gaps can be used as a good estimation for
        # excitation energies. Current implementation takes the second choice so
        # as to make the TDDFT excitation energies agree to DFT orbital energy gaps.
        #
        # There might be a third treatment: Taking treatment 1 first then adding
        # certain types of corrections to the excitation energy at last.
        # I'm not sure how to do this properly.
        #
        # See also issue https://github.com/pyscf/pyscf/issues/1187

        vind, hdiag = rhf.TDA.gen_vind(self, mf)
        def vindp(x):
            with lib.temporary_env(mf, exxdiv=None):
                return vind(x)
        return vindp, hdiag

    def get_ab(self, mf=None):
        raise NotImplementedError

    def nuc_grad_method(self):
        raise NotImplementedError

CIS = TDA


class TDHF(rhf.TDHF):
    def gen_vind(self, mf):
        vind, hdiag = rhf.TDHF.gen_vind(self, mf)
        def vindp(x):
            with lib.temporary_env(mf, exxdiv=None):
                return vind(x)
        return vindp, hdiag

    def get_ab(self, mf=None):
        raise NotImplementedError

    def nuc_grad_method(self):
        raise NotImplementedError

RPA = TDRHF = TDHF


from pyscf.pbc import scf
scf.hf.RHF.TDA = lib.class_as_method(TDA)
scf.hf.RHF.TDHF = lib.class_as_method(TDHF)
scf.rohf.ROHF.TDA = None
scf.rohf.ROHF.TDHF = None




#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
Numerical integration functions for 4-component or 2-component relativistic
methods with j-adapted AO basis
'''

import numpy
from pyscf import lib
from pyscf.dft import numint
from pyscf.dft.numint import _dot_ao_dm, _dot_ao_ao, _scale_ao, _tau_dot, BLKSIZE
from pyscf.dft.numint2c import _eval_xc_eff, mcfun_eval_xc_adapter
from pyscf.dft import xc_deriv
from pyscf import __config__


def eval_ao(mol, coords, deriv=0, with_s=True, shls_slice=None,
            non0tab=None, cutoff=None, out=None, verbose=None):
    '''Evaluates the value of 2-component or 4-component j-adapted basis on grids.
    '''
    comp = (deriv+1)*(deriv+2)*(deriv+3)//6
    feval = 'GTOval_spinor_deriv%d' % deriv
    if not with_s:
        # aoLa, aoLb = aoL
        ao = aoL = mol.eval_gto(feval, coords, comp, shls_slice, non0tab,
                                cutoff=cutoff, out=out)
    else:
        assert (deriv <= 1)  # only GTOval_ipsp_spinor
        ngrids = coords.shape[0]
        nao = mol.nao_2c()
        ao = numpy.ndarray((4,comp,nao,ngrids), dtype=numpy.complex128, buffer=out)
        aoL = mol.eval_gto(feval, coords, comp, shls_slice, non0tab, out=ao[:2])  # noqa
        ao = ao.transpose(0,1,3,2)
        aoS = ao[2:]
        aoSa, aoSb = aoS
        feval_gto = ['GTOval_sp_spinor', 'GTOval_ipsp_spinor']
        p1 = 0
        for n in range(deriv+1):
            comp = (n+1)*(n+2)//2
            p0, p1 = p1, p1 + comp
            aoSa[p0:p1], aoSb[p0:p1] = mol.eval_gto(
                feval_gto[n], coords, comp, shls_slice, non0tab, cutoff=cutoff)

        if deriv == 0:
            ao = ao[:,0]
    return ao

def _dot_spinor_dm(mol, ket, dm, non0tab, shls_slice, ao_loc):
    ket_a, ket_b = ket
    outa = _dot_ao_dm(mol, ket_a, dm, non0tab, shls_slice, ao_loc)
    outb = _dot_ao_dm(mol, ket_b, dm, non0tab, shls_slice, ao_loc)
    return outa, outb

def _contract_rho_m(bra, ket, hermi=0, bra_eq_ket=False):
    '''
    hermi indicates whether the density matrix is hermitian.
    bra_eq_ket indicates whether bra and ket basis are the same AOs.
    '''
    ket_a, ket_b = ket
    bra_a, bra_b = bra
    ngrids = ket_a.shape[0]
    if hermi:
        raa = numpy.einsum('pi,pi->p', bra_a.real, ket_a.real)
        raa+= numpy.einsum('pi,pi->p', bra_a.imag, ket_a.imag)
        rab = numpy.einsum('pi,pi->p', bra_a.conj(), ket_b)
        rbb = numpy.einsum('pi,pi->p', bra_b.real, ket_b.real)
        rbb+= numpy.einsum('pi,pi->p', bra_b.imag, ket_b.imag)
        rho_m = numpy.empty((4, ngrids))
        rho_m[0,:] = raa + rbb     # rho
        rho_m[1,:] = rab.real      # mx
        rho_m[2,:] = rab.imag      # my
        rho_m[3,:] = raa - rbb     # mz
        if bra_eq_ket:
            rho_m[1,:] *= 2
            rho_m[2,:] *= 2
        else:
            rba = numpy.einsum('pi,pi->p', bra_b.conj(), ket_a)
            rho_m[1,:] += rba.real
            rho_m[2,:] -= rba.imag
    else:
        raa = numpy.einsum('pi,pi->p', bra_a.conj(), ket_a)
        rab = numpy.einsum('pi,pi->p', bra_a.conj(), ket_b)
        rba = numpy.einsum('pi,pi->p', bra_b.conj(), ket_a)
        rbb = numpy.einsum('pi,pi->p', bra_b.conj(), ket_b)
        rho_m = numpy.empty((4, ngrids), dtype=numpy.complex128)
        rho_m[0,:] = raa + rbb         # rho
        rho_m[1,:] = rba + rab         # mx
        rho_m[2,:] = (rba - rab) * 1j  # my
        rho_m[3,:] = raa - rbb         # mz
    return rho_m

def _eval_rho_2c(mol, ao, dm, non0tab=None, xctype='LDA', hermi=0, with_lapl=False):
    aoa, aob = ao
    ngrids, nao = aoa.shape[-2:]
    xctype = xctype.upper()

    if non0tab is None:
        non0tab = numpy.ones(((ngrids+BLKSIZE-1)//BLKSIZE,mol.nbas),
                             dtype=numpy.uint8)
    shls_slice = (0, mol.nbas)
    ao_loc = mol.ao_loc_2c()

    if xctype == 'LDA':
        c0 = _dot_spinor_dm(mol, ao, dm, non0tab, shls_slice, ao_loc)
        rho_m = _contract_rho_m(ao, c0, hermi, True)
    elif xctype == 'GGA':
        # first 4 ~ (rho, m), second 4 ~ (0th order, dx, dy, dz)
        if hermi:
            rho_m = numpy.empty((4, 4, ngrids))
        else:
            rho_m = numpy.empty((4, 4, ngrids), dtype=numpy.complex128)
        c0 = _dot_spinor_dm(mol, ao[:,0], dm, non0tab, shls_slice, ao_loc)
        rho_m[:,0] = _contract_rho_m(ao[:,0], c0, hermi, True)
        for i in range(1, 4):
            rho_m[:,i] = _contract_rho_m(ao[:,i], c0, hermi, False)
        if hermi:
            rho_m[:,1:4] *= 2  # *2 for |ao> dm < dx ao| + |dx ao> dm < ao|
        else:
            for i in range(1, 4):
                c1 = _dot_spinor_dm(mol, ao[:,i], dm, non0tab, shls_slice, ao_loc)
                rho_m[:,i] += _contract_rho_m(ao[:,0], c1, hermi, False)
    else: # meta-GGA
        if hermi:
            dtype = numpy.double
        else:
            dtype = numpy.complex128
        if with_lapl:
            rho_m = numpy.empty((4, 6, ngrids), dtype=dtype)
            tau_idx = 5
        else:
            rho_m = numpy.empty((4, 5, ngrids), dtype=dtype)
            tau_idx = 4
        c0 = _dot_spinor_dm(mol, ao[:,0], dm, non0tab, shls_slice, ao_loc)
        rho_m[:,0] = _contract_rho_m(ao[:,0], c0, hermi, True)
        rho_m[:,tau_idx] = 0
        for i in range(1, 4):
            c1 = _dot_spinor_dm(mol, ao[:,i], dm, non0tab, shls_slice, ao_loc)
            rho_m[:,tau_idx] += _contract_rho_m(ao[:,i], c1, hermi, True)
            rho_m[:,i] = _contract_rho_m(ao[:,i], c0, hermi, False)
            if hermi:
                rho_m[:,i] *= 2
            else:
                rho_m[:,i] += _contract_rho_m(ao[:,0], c1, hermi, False)
        if with_lapl:
            raise NotImplementedError
        # tau = 1/2 (\nabla f)^2
        rho_m[:,tau_idx] *= .5
    return rho_m

def eval_rho(mol, ao, dm, non0tab=None, xctype='LDA', hermi=0, with_lapl=False,
             verbose=None):
    r'''Calculate the electron density and magnetization spin density with
    j-adapted spinor basis.
    ''' + numint.eval_rho.__doc__
    with_s = len(ao) == 4  # aoLa, aoLb, aoSa, aoSb
    if with_s:
        n2c = dm.shape[1] // 2
        dmLL = dm[:n2c,:n2c].copy()
        dmSS = dm[n2c:,n2c:].copy()
        c1 = .5 / lib.param.LIGHT_SPEED
        rho  = _eval_rho_2c(mol, ao[:2], dmLL, non0tab, xctype, hermi, with_lapl)
        rhoS = _eval_rho_2c(mol, ao[2:], dmSS, non0tab, xctype, hermi, with_lapl)
        rhoS *= c1**2
        rho[0] += rhoS[0]
        # M = |\beta\Sigma|
        rho[1:4] -= rhoS[1:4]
        return rho
    else:
        return _eval_rho_2c(mol, ao, dm, non0tab, xctype, hermi, with_lapl)

def _ncol_lda_vxc_mat(mol, ao, weight, rho_tm, vxc, mask, shls_slice, ao_loc,
                      hermi, on_LL=True):
    '''Vxc matrix of non-collinear LDA'''
    # NOTE vxc in u/d representation
    aoa, aob = ao
    r, mx, my, mz = rho_tm
    vxc = xc_deriv.ud2ts(vxc)
    vr, vs = vxc[:,0]
    s = lib.norm(rho_tm[1:4], axis=0)

    wv = weight * vr
    with numpy.errstate(divide='ignore',invalid='ignore'):
        ws = vs * weight / s
    ws[s < 1e-20] = 0
    if not on_LL:
        # flip the sign for small components because
        # M = \beta\Sigma = [sigma,  0    ]
        #                   [0    , -sigma]
        ws *= -1

    # * .5 because of v+v.conj().T in r_vxc
    if hermi:
        wv *= .5
        ws *= .5

    # einsum('g,g,xgi,xgj->ij', vr, weight, ao, ao)
    # + einsum('xy,g,g,xgi,ygj->ij', sx, vs*m[0]/s, weight, ao, ao)
    # + einsum('xy,g,g,xgi,ygj->ij', sy, vs*m[1]/s, weight, ao, ao)
    # + einsum('xy,g,g,xgi,ygj->ij', sz, vs*m[2]/s, weight, ao, ao)
    aow = None
    aow = _scale_ao(aoa, ws*(mx+my*1j), out=aow)
    mat = _dot_ao_ao(mol, aob, aow, mask, shls_slice, ao_loc)
    if hermi:
        mat = mat + mat.conj().T
    else:
        aow = _scale_ao(aob, ws*(mx-my*1j), out=aow)  # Mx
        mat+= _dot_ao_ao(mol, aoa, aow, mask, shls_slice, ao_loc)
    aow = _scale_ao(aoa, wv+ws*mz, out=aow)  # Mz
    mat+= _dot_ao_ao(mol, aoa, aow, mask, shls_slice, ao_loc)
    aow = _scale_ao(aob, wv-ws*mz, out=aow)  # Mz
    mat+= _dot_ao_ao(mol, aob, aow, mask, shls_slice, ao_loc)
    return mat

def _ncol_gga_vxc_mat(mol, ao, weight, rho, vxc, mask, shls_slice, ao_loc,
                      hermi, on_LL=True):
    '''Vxc matrix of non-collinear GGA'''
    raise NotImplementedError

def _ncol_mgga_vxc_mat(mol, ao, weight, rho, vxc, mask, shls_slice, ao_loc,
                       hermi, on_LL=True):
    '''Vxc matrix of non-collinear MGGA'''
    raise NotImplementedError

def _col_lda_vxc_mat(mol, ao, weight, rho, vxc, mask, shls_slice, ao_loc,
                     hermi, on_LL=True):
    '''Vxc matrix of collinear LDA'''
    # NOTE vxc in u/d representation
    aoa, aob = ao
    wv = weight * vxc
    if hermi:
        wv *= .5
    if on_LL:
        wva, wvb = wv
    else:  # for SS block
        # v_rho = (vxc_a + vxc_b) * .5
        # v_mz  = (vxc_a - vxc_b) * .5
        # For small components, M = \beta\Sigma leads to
        # (v_rho - sigma_z*v_mz) = [vxc_b, 0    ]
        #                          [0    , vxc_a]
        wvb, wva = wv
    mat  = _dot_ao_ao(mol, aoa, _scale_ao(aoa, wva[0]), mask, shls_slice, ao_loc)
    mat += _dot_ao_ao(mol, aob, _scale_ao(aob, wvb[0]), mask, shls_slice, ao_loc)
    return mat

def _col_gga_vxc_mat(mol, ao, weight, rho, vxc, mask, shls_slice, ao_loc,
                     hermi, on_LL=True):
    '''Vxc matrix of collinear GGA'''
    # NOTE vxc in u/d representation
    aoa, aob = ao
    wv = weight * vxc
    if hermi:
        wv[:,0] *= .5
    if on_LL:
        wva, wvb = wv
    else:  # for SS block
        wvb, wva = wv
    aow = None
    aow = _scale_ao(aoa[:4], wva[:4], out=aow)
    mat = _dot_ao_ao(mol, aoa[0], aow, mask, shls_slice, ao_loc)
    aow = _scale_ao(aob[:4], wvb[:4], out=aow)
    mat+= _dot_ao_ao(mol, aob[0], aow, mask, shls_slice, ao_loc)
    if hermi != 1:
        aow = _scale_ao(aoa[1:4], wva[1:4].conj(), out=aow)
        mat+= _dot_ao_ao(mol, aow, aoa[0], mask, shls_slice, ao_loc)
        aow = _scale_ao(aob[1:4], wvb[1:4].conj(), out=aow)
        mat+= _dot_ao_ao(mol, aow, aob[0], mask, shls_slice, ao_loc)
    return mat

def _col_mgga_vxc_mat(mol, ao, weight, rho, vxc, mask, shls_slice, ao_loc,
                      hermi, on_LL=True):
    '''Vxc matrix of collinear MGGA'''
    # NOTE vxc in u/d representation
    aoa, aob = ao
    wv = weight * vxc
    tau_idx = 4
    wv[:,tau_idx] *= .5  # *.5 for 1/2 in tau
    if hermi:
        wv[:,0] *= .5
        wv[:,tau_idx] *= .5
    if on_LL:
        wva, wvb = wv
    else:
        wvb, wva = wv
    aow = None
    aow = _scale_ao(aoa[:4], wva[:4], out=aow)
    mat = _dot_ao_ao(mol, aoa[0], aow, mask, shls_slice, ao_loc)
    aow = _scale_ao(aob[:4], wvb[:4], out=aow)
    mat+= _dot_ao_ao(mol, aob[0], aow, mask, shls_slice, ao_loc)
    if hermi != 1:
        aow = _scale_ao(aoa[1:4], wva[1:4].conj(), out=aow)
        mat += _dot_ao_ao(mol, aow, aoa[0], mask, shls_slice, ao_loc)
        aow = _scale_ao(aob[1:4], wvb[1:4].conj(), out=aow)
        mat += _dot_ao_ao(mol, aow, aob[0], mask, shls_slice, ao_loc)

    mat += _tau_dot(mol, aoa, aoa, wva[tau_idx], mask, shls_slice, ao_loc)
    mat += _tau_dot(mol, aob, aob, wvb[tau_idx], mask, shls_slice, ao_loc)
    return mat

def _mcol_lda_vxc_mat(mol, ao, weight, rho, vxc, mask, shls_slice, ao_loc,
                      hermi, on_LL=True):
    '''Vxc matrix of multi-collinear LDA'''
    # NOTE vxc in t/m representation
    aoa, aob = ao
    wv = weight * vxc
    if hermi:
        wv *= .5  # * .5 because of v+v.conj().T in r_vxc
    if not on_LL:
        # flip the sign for small components because
        # M = \beta\Sigma = [sigma,  0    ]
        #                   [0    , -sigma]
        wv[1:4] *= -1
    wr, wmx, wmy, wmz = wv

    # einsum('g,g,xgi,xgj->ij', vxc, weight, ao, ao)
    # + einsum('xy,g,g,xgi,ygj->ij', sx, vxc, weight, ao, ao)
    # + einsum('xy,g,g,xgi,ygj->ij', sy, vxc, weight, ao, ao)
    # + einsum('xy,g,g,xgi,ygj->ij', sz, vxc, weight, ao, ao)
    aow = None
    aow = _scale_ao(aoa, wmx[0]+wmy[0]*1j, out=aow)
    mat = _dot_ao_ao(mol, aob, aow, mask, shls_slice, ao_loc)
    if hermi:
        assert vxc.dtype == numpy.double
        mat = mat + mat.conj().T
    else:
        aow = _scale_ao(aob, wmx[0]-wmy[0]*1j, out=aow)
        mat+= _dot_ao_ao(mol, aoa, aow, mask, shls_slice, ao_loc)
    aow = _scale_ao(aoa, wr[0]+wmz[0], out=aow)  # Mz
    mat+= _dot_ao_ao(mol, aoa, aow, mask, shls_slice, ao_loc)
    aow = _scale_ao(aob, wr[0]-wmz[0], out=aow)  # Mz
    mat+= _dot_ao_ao(mol, aob, aow, mask, shls_slice, ao_loc)
    return mat

def _mcol_gga_vxc_mat(mol, ao, weight, rho, vxc, mask, shls_slice, ao_loc,
                      hermi, on_LL=True):
    '''Vxc matrix of multi-collinear LDA'''
    # NOTE vxc in t/m representation
    aoa, aob = ao
    wv = weight * vxc
    if hermi:
        wv[:,0] *= .5
    if not on_LL:
        # flip the sign for small components because
        # M = \beta\Sigma = [sigma,  0    ]
        #                   [0    , -sigma]
        wv[1:4] *= -1
    wr, wmx, wmy, wmz = wv

    aow = None
    aow = _scale_ao(aoa[:4], wmx[:4]+wmy[:4]*1j, out=aow)
    mat = _dot_ao_ao(mol, aob[0], aow, mask, shls_slice, ao_loc)
    if hermi:
        assert vxc.dtype == numpy.double
        mat = mat + mat.conj().T
    else:
        aow = _scale_ao(aob[:4], wmx[:4]-wmy[:4]*1j, out=aow)
        mat+= _dot_ao_ao(mol, aoa[0], aow, mask, shls_slice, ao_loc)
        aow = _scale_ao(aob[1:4], (wmx[1:4]+wmy[1:4]*1j).conj(), out=aow)
        mat+= _dot_ao_ao(mol, aow, aoa[0], mask, shls_slice, ao_loc)
        aow = _scale_ao(aoa[1:4], (wmx[1:4]-wmy[1:4]*1j).conj(), out=aow)
        mat+= _dot_ao_ao(mol, aow, aob[0], mask, shls_slice, ao_loc)
        aow = _scale_ao(aoa[1:4], (wr[1:4]+wmz[1:4]).conj(), out=aow)  # Mz
        mat+= _dot_ao_ao(mol, aow, aoa[0], mask, shls_slice, ao_loc)
        aow = _scale_ao(aob[1:4], (wr[1:4]-wmz[1:4]).conj(), out=aow)  # Mz
        mat+= _dot_ao_ao(mol, aow, aob[0], mask, shls_slice, ao_loc)
    aow = _scale_ao(aoa, wr[:4]+wmz[:4], out=aow)  # Mz
    mat+= _dot_ao_ao(mol, aoa[0], aow, mask, shls_slice, ao_loc)
    aow = _scale_ao(aob, wr[:4]-wmz[:4], out=aow)  # Mz
    mat+= _dot_ao_ao(mol, aob[0], aow, mask, shls_slice, ao_loc)
    return mat

def _mcol_mgga_vxc_mat(mol, ao, weight, rho, vxc, mask, shls_slice, ao_loc,
                       hermi, on_LL=True):
    '''Vxc matrix of multi-collinear MGGA'''
    # NOTE vxc in t/m representation
    aoa, aob = ao
    wv = weight * vxc
    tau_idx = 4
    wv[:,tau_idx] *= .5  # *.5 for 1/2 in tau
    if hermi:
        wv[:,0] *= .5
        wv[:,tau_idx] *= .5
    if not on_LL:
        # flip the sign for small components because
        # M = \beta\Sigma = [sigma,  0    ]
        #                   [0    , -sigma]
        wv[1:4] *= -1
    wr, wmx, wmy, wmz = wv

    aow = None
    aow = _scale_ao(aoa[:4], wmx[:4]+wmy[:4]*1j, out=aow)
    mat = _dot_ao_ao(mol, aob[0], aow, mask, shls_slice, ao_loc)
    mat += _tau_dot(mol, aob, aoa, wmx[tau_idx]+wmy[tau_idx]*1j, mask,
                    shls_slice, ao_loc)
    if hermi:
        assert vxc.dtype == numpy.double
        mat = mat + mat.conj().T
    else:
        aow = _scale_ao(aob[:4], wmx[:4]-wmy[:4]*1j, out=aow)
        mat+= _dot_ao_ao(mol, aoa[0], aow, mask, shls_slice, ao_loc)
        mat+= _tau_dot(mol, aoa, aob, wmx[tau_idx]-wmy[tau_idx]*1j, mask,
                       shls_slice, ao_loc)

        aow = _scale_ao(aob[1:4], (wmx[1:4]+wmy[1:4]*1j).conj(), out=aow)
        mat+= _dot_ao_ao(mol, aow, aoa[0], mask, shls_slice, ao_loc)
        aow = _scale_ao(aoa[1:4], (wmx[1:4]-wmy[1:4]*1j).conj(), out=aow)
        mat+= _dot_ao_ao(mol, aow, aob[0], mask, shls_slice, ao_loc)
        aow = _scale_ao(aoa[1:4], (wr[1:4]+wmz[1:4]).conj(), out=aow)  # Mz
        mat+= _dot_ao_ao(mol, aow, aoa[0], mask, shls_slice, ao_loc)
        aow = _scale_ao(aob[1:4], (wr[1:4]-wmz[1:4]).conj(), out=aow)  # Mz
        mat+= _dot_ao_ao(mol, aow, aob[0], mask, shls_slice, ao_loc)

    aow = _scale_ao(aoa, wr[:4]+wmz[:4], out=aow)  # Mz
    mat+= _dot_ao_ao(mol, aoa[0], aow, mask, shls_slice, ao_loc)
    mat+= _tau_dot(mol, aoa, aoa, wr[tau_idx]+wmz[tau_idx], mask, shls_slice, ao_loc)
    aow = _scale_ao(aob, wr[:4]-wmz[:4], out=aow)  # Mz
    mat+= _dot_ao_ao(mol, aob[0], aow, mask, shls_slice, ao_loc)
    mat+= _tau_dot(mol, aob, aob, wr[tau_idx]-wmz[tau_idx], mask, shls_slice, ao_loc)
    return mat

def r_vxc(ni, mol, grids, xc_code, dms, spin=0, relativity=1, hermi=1,
          max_memory=2000, verbose=None):
    '''Calculate 2-component or 4-component Vxc matrix in j-adapted basis.
    ''' + numint.nr_rks.__doc__
    xctype = ni._xc_type(xc_code)
    shls_slice = (0, mol.nbas)
    ao_loc = mol.ao_loc_2c()
    n2c = ao_loc[-1]

    make_rho, nset, nao = ni._gen_rho_evaluator(mol, dms, hermi)
    with_s = (nao == n2c*2)  # 4C DM
    nelec = numpy.zeros(nset)
    excsum = numpy.zeros(nset)
    matLL = numpy.zeros((nset,n2c,n2c), dtype=numpy.complex128)
    matSS = numpy.zeros_like(matLL)

    if xctype in ('LDA', 'GGA', 'MGGA'):
        f_eval_mat = {
            ('LDA' , 'c'): (_col_lda_vxc_mat  , 0),
            ('GGA' , 'c'): (_col_gga_vxc_mat  , 1),
            ('MGGA', 'c'): (_col_mgga_vxc_mat , 1),
            ('LDA' , 'n'): (_ncol_lda_vxc_mat , 0),
            ('GGA' , 'n'): (_ncol_gga_vxc_mat , 1),
            ('MGGA', 'n'): (_ncol_mgga_vxc_mat, 1),
            ('LDA' , 'm'): (_mcol_lda_vxc_mat , 0),
            ('GGA' , 'm'): (_mcol_gga_vxc_mat , 1),
            ('MGGA', 'm'): (_mcol_mgga_vxc_mat, 1),
        }
        fmat, ao_deriv = f_eval_mat[(xctype, ni.collinear[0])]

        if ni.collinear[0] == 'm':  # mcol
            eval_xc = ni.mcfun_eval_xc_adapter(xc_code)
        else:
            eval_xc = ni.eval_xc_eff

        for ao, mask, weight, coords \
                in ni.block_loop(mol, grids, nao, ao_deriv, max_memory,
                                 with_s=with_s):
            for i in range(nset):
                rho = make_rho(i, ao, mask, xctype)
                exc, vxc = eval_xc(xc_code, rho, deriv=1, xctype=xctype)[:2]
                if xctype == 'LDA':
                    den = rho[0] * weight
                else:
                    den = rho[0,0] * weight
                nelec[i] += den.sum()
                excsum[i] += numpy.dot(den, exc)

                matLL[i] += fmat(mol, ao[:2], weight, rho, vxc,
                                 mask, shls_slice, ao_loc, hermi, True)
                if with_s:
                    matSS[i] += fmat(mol, ao[2:], weight, rho, vxc,
                                     mask, shls_slice, ao_loc, hermi, False)

        if hermi:
            matLL = matLL + matLL.conj().transpose(0,2,1)
            if with_s:
                matSS = matSS + matSS.conj().transpose(0,2,1)
    elif xctype == 'HF':
        pass
    else:
        raise NotImplementedError(f'r_vxc for functional {xc_code}')

    if with_s:
        matSS *= (.5 / lib.param.LIGHT_SPEED)**2
        vmat = numpy.zeros((nset,nao,nao), dtype=numpy.complex128)
        vmat[:,:n2c,:n2c] = matLL
        vmat[:,n2c:,n2c:] = matSS
    else:
        vmat = matLL

    if nset == 1:
        nelec = nelec[0]
        excsum = excsum[0]
    return nelec, excsum, vmat.reshape(dms.shape)

def _col_rho_tm2ud(rho_tm):
    return numpy.stack([(rho_tm[0] + rho_tm[3]) * .5,
                        (rho_tm[0] - rho_tm[3]) * .5,])

def _col_lda_fxc_mat(mol, ao, weight, rho0, rho1, fxc,
                     mask, shls_slice, ao_loc, hermi, on_LL=True):
    '''Kernel matrix of collinear LDA'''
    rho1 = _col_rho_tm2ud(rho1)
    vxc1 = numpy.einsum('ag,abyg->byg', rho1, fxc[:,0,:])
    return _col_lda_vxc_mat(mol, ao, weight, rho0, vxc1, mask, shls_slice,
                            ao_loc, hermi, on_LL)

def _col_gga_fxc_mat(mol, ao, weight, rho0, rho1, fxc,
                     mask, shls_slice, ao_loc, hermi, on_LL=True):
    '''Kernel matrix of collinear GGA'''
    rho1 = _col_rho_tm2ud(rho1)
    vxc1 = numpy.einsum('axg,axbyg->byg', rho1, fxc)
    return _col_gga_vxc_mat(mol, ao, weight, rho0, vxc1, mask, shls_slice,
                            ao_loc, hermi, on_LL)

def _col_mgga_fxc_mat(mol, ao, weight, rho0, rho1, fxc,
                      mask, shls_slice, ao_loc, hermi, on_LL=True):
    '''Kernel matrix of collinear MGGA'''
    rho1 = _col_rho_tm2ud(rho1)
    vxc1 = numpy.einsum('axg,axbyg->byg', rho1, fxc)
    return _col_mgga_vxc_mat(mol, ao, weight, rho0, vxc1, mask, shls_slice,
                             ao_loc, hermi, on_LL)

def _mcol_lda_fxc_mat(mol, ao, weight, rho0, rho1, fxc,
                      mask, shls_slice, ao_loc, hermi, on_LL=True):
    '''Kernel matrix of multi-collinear LDA'''
    vxc1 = numpy.einsum('ag,abyg->byg', rho1, fxc[:,0,:])
    return _mcol_lda_vxc_mat(mol, ao, weight, rho0, vxc1, mask, shls_slice,
                             ao_loc, hermi, on_LL)

def _mcol_gga_fxc_mat(mol, ao, weight, rho0, rho1, fxc,
                      mask, shls_slice, ao_loc, hermi, on_LL=True):
    '''Kernel matrix of multi-collinear GGA'''
    vxc1 = numpy.einsum('axg,axbyg->byg', rho1, fxc)
    return _mcol_gga_vxc_mat(mol, ao, weight, rho0, vxc1, mask, shls_slice,
                             ao_loc, hermi, on_LL)

def _mcol_mgga_fxc_mat(mol, ao, weight, rho0, rho1, fxc,
                       mask, shls_slice, ao_loc, hermi, on_LL=True):
    '''Kernel matrix of multi-collinear MGGA'''
    vxc1 = numpy.einsum('axg,axbyg->byg', rho1, fxc)
    return _mcol_mgga_vxc_mat(mol, ao, weight, rho0, vxc1, mask, shls_slice,
                              ao_loc, hermi, on_LL)

def r_fxc(ni, mol, grids, xc_code, dm0, dms, spin=0, relativity=1, hermi=0,
          rho0=None, vxc=None, fxc=None, max_memory=2000, verbose=None):
    '''Calculate 2-component or 4-component Vxc matrix in j-adapted basis.
    ''' + numint.nr_rks_fxc.__doc__
    xctype = ni._xc_type(xc_code)
    shls_slice = (0, mol.nbas)
    ao_loc = mol.ao_loc_2c()
    n2c = ao_loc[-1]
    if ni.collinear[0] not in ('c', 'm'):  # col or mcol
        raise NotImplementedError('non-collinear fxc')

    make_rho1, nset, nao = ni._gen_rho_evaluator(mol, dms, hermi)
    with_s = (nao == n2c*2)  # 4C DM

    if rho0 is None and (xctype != 'LDA' or fxc is None):
        make_rho0 = ni._gen_rho_evaluator(mol, dm0, 1)[0]
    else:
        make_rho0 = None

    matLL = numpy.zeros((nset,n2c,n2c), dtype=dms.dtype)
    matSS = numpy.zeros_like(matLL)
    if xctype in ('LDA', 'GGA', 'MGGA'):
        f_eval_mat = {
            ('LDA' , 'c'): (_col_lda_fxc_mat   , 0),
            ('GGA' , 'c'): (_col_gga_fxc_mat   , 1),
            ('MGGA', 'c'): (_col_mgga_fxc_mat  , 1),
            ('LDA' , 'm'): (_mcol_lda_fxc_mat  , 0),
            ('GGA' , 'm'): (_mcol_gga_fxc_mat  , 1),
            ('MGGA', 'm'): (_mcol_mgga_fxc_mat , 1),
        }
        fmat, ao_deriv = f_eval_mat[(xctype, ni.collinear[0])]
        if ni.collinear[0] == 'm':  # mcol
            eval_xc = ni.mcfun_eval_xc_adapter(xc_code)
        else:
            eval_xc = ni.eval_xc_eff

        _rho0 = None
        p1 = 0
        for ao, mask, weight, coords \
                in ni.block_loop(mol, grids, nao, ao_deriv, max_memory,
                                 with_s=with_s):
            p0, p1 = p1, p1 + weight.size
            if fxc is None:
                if rho0 is not None:
                    if xctype == 'LDA':
                        _rho0 = numpy.asarray(rho0[:,p0:p1], order='C')
                    else:
                        _rho0 = numpy.asarray(rho0[:,:,p0:p1], order='C')
                elif make_rho0 is not None:
                    _rho0 = make_rho0(0, ao, mask, xctype)

                _fxc = eval_xc(xc_code, _rho0, deriv=2, xctype=xctype)[2]
            else:
                _fxc = fxc[:,:,:,:,p0:p1]

            for i in range(nset):
                rho1 = make_rho1(i, ao, mask, xctype)
                matLL[i] += fmat(mol, ao[:2], weight, _rho0, rho1, _fxc,
                                 mask, shls_slice, ao_loc, hermi)
                if with_s:
                    matSS[i] += fmat(mol, ao[2:], weight, _rho0, rho1, _fxc,
                                     mask, shls_slice, ao_loc, hermi, False)

        if hermi:
            # for (\nabla\mu) \nu + \mu (\nabla\nu)
            matLL = matLL + matLL.conj().transpose(0,2,1)
            if with_s:
                matSS = matSS + matSS.conj().transpose(0,2,1)

    elif xctype == 'HF':
        pass
    else:
        raise NotImplementedError(f'r_fxc for functional {xc_code}')

    if with_s:
        matSS *= (.5 / lib.param.LIGHT_SPEED)**2
        vmat = numpy.zeros((nset,nao,nao), dtype=numpy.complex128)
        vmat[:,:n2c,:n2c] = matLL
        vmat[:,n2c:,n2c:] = matSS
    else:
        vmat = matLL

    if isinstance(dms, numpy.ndarray) and dms.ndim == 2:
        vmat = vmat[0]
    return vmat

def cache_xc_kernel(ni, mol, grids, xc_code, mo_coeff, mo_occ, spin=1,
                    max_memory=2000):
    '''Compute the 0th order density, Vxc and fxc.  They can be used in TDDFT,
    DFT hessian module etc.
    '''
    xctype = ni._xc_type(xc_code)
    if xctype in ('GGA', 'MGGA'):
        ao_deriv = 1
    elif xctype == 'NLC':
        raise NotImplementedError('NLC')
    else:
        ao_deriv = 0

    # Ignore density laplacian for mcfun
    dm = numpy.dot(mo_coeff * mo_occ, mo_coeff.conj().T)
    hermi = 1
    make_rho, nset, nao = ni._gen_rho_evaluator(mol, dm, hermi)
    n2c = mol.nao_2c()
    with_s = (nao == n2c*2)  # 4C DM
    rho = []
    for ao, mask, weight, coords \
            in ni.block_loop(mol, grids, nao, ao_deriv, max_memory,
                             with_s=with_s):
        rho.append(make_rho(0, ao, mask, xctype))
    rho = numpy.hstack(rho)

    if ni.collinear[0] == 'm':  # mcol
        eval_xc = ni.mcfun_eval_xc_adapter(xc_code)
    else:
        eval_xc = ni.eval_xc_eff
    vxc, fxc = eval_xc(xc_code, rho, deriv=2, xctype=xctype)[1:3]
    return rho, vxc, fxc

def get_rho(ni, mol, dm, grids, max_memory=2000):
    make_rho, nset, nao = ni._gen_rho_evaluator(mol, dm, hermi=1)
    n2c = mol.nao_2c()
    with_s = (nao == n2c*2)  # 4C DM
    rho = numpy.empty(grids.weights.size)
    p1 = 0
    for ao, mask, weight, coords \
            in ni.block_loop(mol, grids, nao, 0, max_memory, with_s=with_s):
        p0, p1 = p1, p1 + weight.size
        rho[p0:p1] = make_rho(0, ao, mask, 'LDA')[0]
    return rho


class RNumInt(numint._NumIntMixin):
    '''NumInt for j-adapted (spinor) basis'''

    # collinear schemes:
    #   'col' (collinear, by default)
    #   'ncol' (non-collinear)
    #   'mcol' (multi-collinear)
    collinear = getattr(__config__, 'dft_numint_RnumInt_collinear', 'col')
    spin_samples = getattr(__config__, 'dft_numint_RnumInt_spin_samples', 770)
    collinear_thrd = getattr(__config__, 'dft_numint_RnumInt_collinear_thrd', 0.99)
    collinear_samples = getattr(__config__, 'dft_numint_RnumInt_collinear_samples', 200)

    get_rho = get_rho
    cache_xc_kernel = cache_xc_kernel
    get_vxc = r_vxc = r_vxc
    get_fxc = r_fxc = r_fxc

    eval_xc_eff = _eval_xc_eff
    mcfun_eval_xc_adapter = mcfun_eval_xc_adapter
    eval_ao = staticmethod(eval_ao)
    eval_rho = staticmethod(eval_rho)

    def eval_rho1(self, mol, ao, dm, screen_index=None, xctype='LDA', hermi=0,
                  with_lapl=True, cutoff=None, ao_cutoff=None, pair_mask=None,
                  verbose=None):
        return self.eval_rho(mol, ao, dm, screen_index, xctype, hermi,
                             with_lapl, verbose=verbose)

    def eval_rho2(self, mol, ao, mo_coeff, mo_occ, non0tab=None, xctype='LDA',
                  with_lapl=True, verbose=None):
        raise NotImplementedError

    def block_loop(self, mol, grids, nao, deriv=0, max_memory=2000,
                   non0tab=None, blksize=None, buf=None, with_s=False):
        '''Define this macro to loop over grids by blocks.
        '''
        if grids.coords is None:
            grids.build(with_non0tab=True)
        ngrids = grids.weights.size
        comp = (deriv+1)*(deriv+2)*(deriv+3)//6
# NOTE to index ni.non0tab, the blksize needs to be the integer multiplier of BLKSIZE
        if blksize is None:
            blksize = min(int(max_memory*1e6/((comp*4+4)*nao*16*BLKSIZE))*BLKSIZE, ngrids)
            blksize = max(blksize, BLKSIZE)
        if non0tab is None:
            non0tab = grids.non0tab
        if non0tab is None:
            non0tab = numpy.ones(((ngrids+BLKSIZE-1)//BLKSIZE,mol.nbas),
                                 dtype=numpy.uint8)

        if buf is None:
            buf = numpy.empty((4,comp,blksize,nao), dtype=numpy.complex128)
        for ip0 in range(0, ngrids, blksize):
            ip1 = min(ngrids, ip0+blksize)
            coords = grids.coords[ip0:ip1]
            weight = grids.weights[ip0:ip1]
            non0 = non0tab[ip0//BLKSIZE:]
            ao = self.eval_ao(mol, coords, deriv=deriv, with_s=with_s,
                              non0tab=non0, cutoff=grids.cutoff, out=buf)
            yield ao, non0, weight, coords

    def _gen_rho_evaluator(self, mol, dms, hermi=1, with_lapl=False, grids=None):
        dms = numpy.asarray(dms, order='C')
        if dms.ndim == 2:
            dms = dms[numpy.newaxis]
        ndms, nao = dms.shape[:2]
        def make_rho(idm, ao, non0tab, xctype):
            return self.eval_rho(mol, ao, dms[idm], non0tab, xctype, hermi, with_lapl)
        return make_rho, ndms, nao

_RNumInt = RNumInt



#!/usr/bin/env python
# Copyright 2014-2021 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Timothy Berkelbach <tim.berkelbach@gmail.com>
#

'''
Restricted QCISD implementation
The 4-index integrals are saved on disk entirely (without using any symmetry).

Note MO integrals are treated in chemist's notation

Ref:
'''


import numpy
import numpy as np

from pyscf import lib
from pyscf.lib import logger
from pyscf.cc import rccsd_slow as rccsd
from pyscf.cc import rintermediates as imd
from pyscf import __config__

BLKMIN = getattr(__config__, 'cc_ccsd_blkmin', 4)
MEMORYMIN = getattr(__config__, 'cc_ccsd_memorymin', 2000)


def kernel(mycc, eris=None, t1=None, t2=None, max_cycle=50, tol=1e-8,
           tolnormt=1e-6, verbose=None):
    '''Same as ccsd.kernel with strings modified to correct the method name'''
    log = logger.new_logger(mycc, verbose)
    if eris is None:
        eris = mycc.ao2mo(mycc.mo_coeff)
    if t1 is None and t2 is None:
        t1, t2 = mycc.get_init_guess(eris)
    elif t2 is None:
        t2 = mycc.get_init_guess(eris)[1]

    cput1 = cput0 = (logger.process_clock(), logger.perf_counter())
    eold = 0
    eccsd = mycc.energy(t1, t2, eris)
    log.info('Init E_corr(QCISD) = %.15g', eccsd)

    if isinstance(mycc.diis, lib.diis.DIIS):
        adiis = mycc.diis
    elif mycc.diis:
        adiis = lib.diis.DIIS(mycc, mycc.diis_file, incore=mycc.incore_complete)
        adiis.space = mycc.diis_space
    else:
        adiis = None

    conv = False
    for istep in range(max_cycle):
        t1new, t2new = mycc.update_amps(t1, t2, eris)
        tmpvec = mycc.amplitudes_to_vector(t1new, t2new)
        tmpvec -= mycc.amplitudes_to_vector(t1, t2)
        normt = numpy.linalg.norm(tmpvec)
        tmpvec = None
        if mycc.iterative_damping < 1.0:
            alpha = mycc.iterative_damping
            t1new = (1-alpha) * t1 + alpha * t1new
            t2new *= alpha
            t2new += (1-alpha) * t2
        t1, t2 = t1new, t2new
        t1new = t2new = None
        t1, t2 = mycc.run_diis(t1, t2, istep, normt, eccsd-eold, adiis)
        eold, eccsd = eccsd, mycc.energy(t1, t2, eris)
        log.info('cycle = %d  E_corr(QCISD) = %.15g  dE = %.9g  norm(t1,t2) = %.6g',
                 istep+1, eccsd, eccsd - eold, normt)
        cput1 = log.timer('QCISD iter', *cput1)
        if abs(eccsd-eold) < tol and normt < tolnormt:
            conv = True
            break
    log.timer('QCISD', *cput0)
    return conv, eccsd, t1, t2


def update_amps(cc, t1, t2, eris):
    # Ref: Hirata et al., J. Chem. Phys. 120, 2581 (2004) Eqs.(35)-(36)
    nocc, nvir = t1.shape
    fock = eris.fock

    fov = fock[:nocc,nocc:].copy()
    foo = fock[:nocc,:nocc].copy()
    fvv = fock[nocc:,nocc:].copy()

    Foo = imd.cc_Foo(0*t1,t2,eris)
    Fvv = imd.cc_Fvv(0*t1,t2,eris)
    Fov = imd.cc_Fov(t1,t2,eris)

    Foo -= np.diag(np.diag(foo))
    Fvv -= np.diag(np.diag(fvv))

    # T1 equation
    t1new = np.asarray(fov).conj().copy()
    t1new +=   lib.einsum('ac,ic->ia', Fvv, t1)
    t1new +=  -lib.einsum('ki,ka->ia', Foo, t1)
    t1new += 2*lib.einsum('kc,kica->ia', Fov, t2)
    t1new +=  -lib.einsum('kc,ikca->ia', Fov, t2)
    t1new += 2*lib.einsum('kcai,kc->ia', eris.ovvo, t1)
    t1new +=  -lib.einsum('kiac,kc->ia', eris.oovv, t1)
    eris_ovvv = np.asarray(eris.ovvv)
    t1new += 2*lib.einsum('kdac,ikcd->ia', eris_ovvv, t2)
    t1new +=  -lib.einsum('kcad,ikcd->ia', eris_ovvv, t2)
    t1new +=-2*lib.einsum('kilc,klac->ia', eris.ooov, t2)
    t1new +=   lib.einsum('likc,klac->ia', eris.ooov, t2)

    # T2 equation
    t2new = np.asarray(eris.ovov).conj().transpose(0,2,1,3).copy()
    Loo = imd.Loo(0*t1, t2, eris)
    Lvv = imd.Lvv(0*t1, t2, eris)
    Loo -= np.diag(np.diag(foo))
    Lvv -= np.diag(np.diag(fvv))
    Woooo = imd.cc_Woooo(0*t1, t2, eris)
    Wvoov = imd.cc_Wvoov(0*t1, t2, eris)
    Wvovo = imd.cc_Wvovo(0*t1, t2, eris)
    Wvvvv = imd.cc_Wvvvv(0*t1, t2, eris)
    t2new += lib.einsum('klij,klab->ijab', Woooo, t2)
    t2new += lib.einsum('abcd,ijcd->ijab', Wvvvv, t2)
    tmp = lib.einsum('ac,ijcb->ijab', Lvv, t2)
    t2new += (tmp + tmp.transpose(1,0,3,2))
    tmp = lib.einsum('ki,kjab->ijab', Loo, t2)
    t2new -= (tmp + tmp.transpose(1,0,3,2))
    tmp  = 2*lib.einsum('akic,kjcb->ijab', Wvoov, t2)
    tmp -=   lib.einsum('akci,kjcb->ijab', Wvovo, t2)
    t2new += (tmp + tmp.transpose(1,0,3,2))
    tmp = lib.einsum('akic,kjbc->ijab', Wvoov, t2)
    t2new -= (tmp + tmp.transpose(1,0,3,2))
    tmp = lib.einsum('bkci,kjac->ijab', Wvovo, t2)
    t2new -= (tmp + tmp.transpose(1,0,3,2))

    tmp2 = np.asarray(eris.ovvv).conj().transpose(1,3,0,2)
    tmp = lib.einsum('abic,jc->ijab', tmp2, t1)
    t2new += (tmp + tmp.transpose(1,0,3,2))
    tmp2 = np.asarray(eris.ooov).transpose(3,1,2,0).conj()
    tmp = lib.einsum('akij,kb->ijab', tmp2, t1)
    t2new -= (tmp + tmp.transpose(1,0,3,2))

    mo_e = eris.fock.diagonal().real
    eia = mo_e[:nocc,None] - mo_e[None,nocc:]
    eijab = lib.direct_sum('ia,jb->ijab',eia,eia)
    t1new /= eia
    t2new /= eijab

    return t1new, t2new


class QCISD(rccsd.RCCSD):
    '''restricted QCISD
    '''

    def kernel(self, t1=None, t2=None, eris=None):
        return self.qcisd(t1, t2, eris)
    def qcisd(self, t1=None, t2=None, eris=None):
        assert (self.mo_coeff is not None)
        assert (self.mo_occ is not None)

        if self.verbose >= logger.WARN:
            self.check_sanity()
        self.dump_flags()

        self.e_hf = self.get_e_hf()

        if eris is None:
            eris = self.ao2mo(self.mo_coeff)

        self.converged, self.e_corr, self.t1, self.t2 = \
                kernel(self, eris, t1, t2, max_cycle=self.max_cycle,
                       tol=self.conv_tol, tolnormt=self.conv_tol_normt,
                       verbose=self.verbose)
        self._finalize()
        return self.e_corr, self.t1, self.t2

    def energy(self, t1=None, t2=None, eris=None):
        return rccsd.energy(self, t1*0, t2, eris)

    update_amps = update_amps

    def qcisd_t(self, t1=None, t2=None, eris=None):
        from pyscf.cc import qcisd_t_slow as qcisd_t
        if t1 is None: t1 = self.t1
        if t2 is None: t2 = self.t2
        if eris is None: eris = self.ao2mo(self.mo_coeff)
        return qcisd_t.kernel(self, eris, t1, t2, self.verbose)

    def density_fit(self, auxbasis=None, with_df=None):
        raise NotImplementedError


if __name__ == '__main__':
    from pyscf import gto, scf

    mol = gto.Mole()
    mol.atom = """C  0.000  0.000  0.000
                  H  0.637  0.637  0.637
                  H -0.637 -0.637  0.637
                  H -0.637  0.637 -0.637
                  H  0.637 -0.637 -0.637"""
    mol.basis = 'cc-pvdz'
    mol.verbose = 7
    mol.spin = 0
    mol.build()
    mf = scf.RHF(mol).run(conv_tol=1e-14)

    mycc = QCISD(mf, frozen=1)
    ecc, t1, t2 = mycc.kernel()
    print(mycc.e_tot - -40.383989)
    et = mycc.qcisd_t()
    print(mycc.e_tot+et - -40.387679)
